/** @preserve Eloquence WEBDLG v2
 (C) Copyright Marxmeier Software AG, 2020-2021
 Version: 1.966 2021-09-24
 $Id: 00-base.js,v 28.54 2021/09/24 14:32:55 mike Exp $
*//*======================================================================
   WEBDLG v2 namespace / globals
========================================================================*/

var eq = {
   c : {},           // Classes
   p : {             // Properties:
      version : 1966 //  Version 1.966 (major*1,000 + minor)
   }
};

/*========================================================================
   class Obj
   Generic base class.
========================================================================*/

eq.c.Obj = function() {};

/*------------------------------------------------------------------------
   static Obj.subClass()
   Subclassing convenience helper.
------------------------------------------------------------------------*/

(eq.c.Obj.setClassName = function(name, cls) {
   var pd = Object.getOwnPropertyDescriptor(cls.prototype.constructor, 'name');
   if (pd !== undefined)
      pd.value = name;
   else
      pd = {
         value : name,
         configurable : true,
         enumerable : false,
         writable : false
      };
   Object.defineProperty(cls.prototype.constructor, 'name', pd);
})("Obj", eq.c.Obj);

eq.c.Obj.subClass = function(name, cls) {
   var base = this;
   cls.prototype = Object.create(base.prototype);
   Object.defineProperty(cls.prototype, 'constructor', {
      value : cls,
      configurable : false,
      enumerable : false,
      writable : true
   });
   eq.c.Obj.setClassName(name, cls);
   cls.subClass = function(n, c) { return base.subClass.call(this, n, c) };
   return eq.c[name] = cls;
};

/** @preserve $Id: 05-const.js,v 28.37 2021/09/23 13:23:37 mike Exp $
*//*======================================================================
   static MainOp
   API->Main message opcodes
========================================================================*/

eq.MainOp = {
   ssnNewOrResume     : 1,  // Data: session id
   applicationStopped : 2,  // Data: restartable true/false
   applicationFailed  : 3,  // Data: none
   layout             : 4,  // Data: layout mode
   cancelCurrentCall  : 5,  // Data: none
   resume             : 6,  // Data: none
   addControl         : 7,  // Data: dialog id, DOM element template
   modifyControl      : 8,  // Data: dialog id, DOM element template
   deleteControl      : 9,  // Data: dialog id, element id
   invalidate         : 10, // Data: dialog id array
   dlg_DRAW           : 11, // Data: dialog id
   dlg_DO             : 12, // Data: dialog id, focus element id
   dlg_POPUP_BOX      : 13, // Data: POPUP BOX descriptor
   openUrl            : 14, // Data: URL, window name
   startApp           : 15, // Data: application name
   activate           : 16, // Data: dialog title
   tabOrderUpdated    : 17, // Data: dialog id, tab order array
   notify             : 18  // Data: dialog id, rule
};

/*========================================================================
   static ApiOp
   Main->API message opcodes
========================================================================*/

eq.ApiOp = {
   documentComplete : 1, // Data: layout grid
   openWebSocket    : 2, // Data: WebSocket URL, session id
   onEvent          : 3, // Data: element id, event type
   onPopupEvent     : 4, // Data: button number
   onDraw           : 5, // Data: none
   onDo             : 6, // Data: dialog id
   onHelp           : 7, // Data: element id
   invalidate       : 8, // Data: dialog id
   updateTabOrder   : 9  // Data: dialog id
};

/*========================================================================
   static DlgRq.tag
   DLG request tags.
========================================================================*/

eq.RqTag  = {
   child : 1, // Child control

   // Global properties

   baseUrl     : 2, // Base url
   helpBaseUrl : 3, // Help base url
   layout      : 4, // Dialog layout mode
   openUrl     : 5, // Open url
   openHelpUrl : 6, // Open help url
   startApp    : 7, // Start application
   activate    : 8, // Activate window

   // Common attributes.

   cssid       : 10, // CSS id
   cssclsDel   : 11, // Delete CSS class(es)
   cssclsAdd   : 12, // Add CSS class(es)
   raster      : 13, // Raster
   posSize     : 14, // Position/size
   size        : 15, // Size, ToolBar children
   border      : 16, // Border
   rule        : 17, // Rule
   ruleMode    : 18, // When to submit rule (Tree/ListBox.singleclick)
   tabonenter  : 19, // Enter key navigates focus
   visible     : 20, // Visibility
   sensitive   : 21, // Sensitivity
   inactive    : 22, // Inactive, eg. TabBox tab
   tabStop     : 23, // Tab stop
   bgColor     : 24, // Background color
   fgColor     : 25, // Foreground color
   url         : 26, // Url
   help        : 27, // Help url

   // Dialog attributes.

   dialogState   : 30, // Resizepolicy, maximized
   dialogPosSize : 31, // Dialog position/size
   fKey          : 32, // Function keys
   dialogDo      : 33, // Dialog.do

   // GroupBox attributes.

   vWidth  : 40, // Visible width
   vHeight : 41, // Visible height

   // Button, Menu, StaticText attributes.

   text       : 50, // Text
   textPos    : 51, // Text position
   align      : 52, // Align
   active     : 53, // Active
   radioGroup : 54, // RadioButton group
   icon       : 55, // PushButton icon
   iconActive : 56, // PushButton icon when active

   // EditText attributes.

   multiline  : 60, // Multiline
   readonly   : 61, // Read-only
   content    : 62, // Content
   ispassword : 63, // IsPassword

   // ListBox, PopText, Tree attributes.

   columns    : 70, // Columns
   activeLine : 71, // Active line
   topItem    : 72, // First visible line
   status     : 73, // Status

   // TabBox attributes.

   tabs        : 80, // Tabs
   selectedTab : 81, // Selected tab

   // Splitter, ProgressBar attributes.

   panes       : 90, // Setup splitter panes
   vertical    : 91, // Split vertical
   position    : 92, // Divider position
   quickExpand : 93, // Quick-expand controls

   // ComboBox attributes.

   items   : 100, // List items
   boxSize : 101, // Box width/height

   // ProgressBar attributes.

   value : 110  // Value, empty=indeterminate
};

/*========================================================================
   static RsValType
   Response changed element value type.
========================================================================*/

eq.RsValType = {
   idFocus       : 1,  // Focus object id, STRING
   ruleOnce      : 2,  // Submit this rule once, INT
   textValue     : 3,  // Text value, STRING
   buttonChecked : 4,  // Button checked, UINT
   caretPos      : 5,  // Caret position, UINT
   activeLine    : 6,  // Active line, UINT
   topItem       : 7,  // First visible line, UINT
   selectedTab   : 8,  // Selected tab, UINT
   splitterPos   : 9,  // Splitter position, UINT
   status        : 10, // ListBox/Tree status, ARRAY
   ruleKey       : 11, // Rule submission key code, UINT
   mouseBtns     : 12, //                 mouse buttons, UINT
   mouseCnt      : 13, //                 mouse click count, UINT
   mouseMods     : 14, //                 mouse modifier keys, UINT
   notified      : 15, // Flag: Dialog notified
   dialogMax     : 16, // Dialog maximized, UINT
   dialogPos     : 17, // Dialog position, INT x/y
   dialogSize    : 18  // Dialog size, INT w/h
};

/*========================================================================
   static LayoutMode
   Layout mode constants.
========================================================================*/

eq.layoutMode = {
   inline    : 1,
   maximized : 2,
   dialog    : 4
};

/*========================================================================
   static DialogMode, DialogState
   Dialog mode and state constants.
========================================================================*/

eq.DialogMode = {
   none   : 0,
   layout : 1,
   scale  : 2,
   inline : 3
};

eq.DialogState = {
   maximized : 1
};

/** @preserve $Id: 05-util.js,v 28.40 2021/09/23 13:24:19 mike Exp $
*//*======================================================================
   static Dom
   DOM utilities.
========================================================================*/

eq.Dom = {};

/*------------------------------------------------------------------------
   static Dom.viewRect()
   Obtain current view rectangle.
------------------------------------------------------------------------*/

eq.Dom.viewRect = function() {
   return {
      x : window.pageXOffset,
      y : window.pageYOffset,
      w : document.documentElement.clientWidth,
      h : document.documentElement.clientHeight
   };
};

/*------------------------------------------------------------------------
   static Dom.setContent()
   Set element text or HTML content.
------------------------------------------------------------------------*/

eq.Dom.setContent = function(el, cn, checkAccel) {
   var self = this.setContent, html = self.isHtml.exec(cn);
   if (html) {
      this.deleteChildren(el);
      this.setHtmlContent(html[1], el);
   }
   else if (checkAccel)
      el.textContent = cn.replace('&', '');
   else
      el.textContent = cn;
};

eq.Dom.setContent.isHtml = new RegExp('^<html>(.*)(</html>)?$', 'i');
eq.Dom.setContent.parseHtml = new DOMParser();

eq.Dom.setHtmlContent = function(cn, body, head) {
   var
      html = this.setContent.parseHtml.parseFromString(cn, 'text/html'),
      from, child;
   if (head && (from = html.head))
      while (child = from.firstChild)
         head.appendChild(child);
   if (body && (from = html.body))
      while (child = from.firstChild)
         body.appendChild(child);
};

/*------------------------------------------------------------------------
   static Dom.copyContent()
   Copy element text or HTML content.
------------------------------------------------------------------------*/

eq.Dom.copyContent = function(to, from) {
   this.deleteChildren(to);
   if (from) {
      var child = from.firstChild;
      while (child) {
         to.appendChild(child.cloneNode(true));
         child = child.nextSibling;
      }
   }
};

/*------------------------------------------------------------------------
   static Dom.updateQueryString()
   Update query string, add/replace or remove parameter.
------------------------------------------------------------------------*/

eq.Dom.updateQueryString = function(url, key, val) {
   var
      r = this.updateQueryString.re.exec(url),
      q = r[2];
   if (q) {
      var re = new RegExp('([?&])' + key + '=[^&#]*', 'i');
      if (q.match(re)) {
         if (val !== undefined)
            q = q.replace(re, '$1' + key + '=' + val);
         else {
            q = q.replace(re, '');
            if (q.length && q.charAt(0) === '&')
               q = q.replace('&', '?');
         }
      }
      else if (val !== undefined)
         q = q + '&' + key + '=' + val;
   }
   else if (val !== undefined)
      q = '?' + key + '=' + val;
   url = r[1];
   if (q)
      url += q;
   if (r[3])
      url += r[3];
   return url;
};

eq.Dom.updateQueryString.re = new RegExp('^([^?#]+)([?][^#]*)?(#.*)?$');

/*------------------------------------------------------------------------
   static Dom.zIndex()
   Return numeric element z-index.
------------------------------------------------------------------------*/

eq.Dom.zIndex = function(el) {
   var z = window.getComputedStyle(el).zIndex;
   return +z === +z ? +z : 0;
};

/*------------------------------------------------------------------------
   static Dom.eqControl()
   Obtain 'eq-control' element.
------------------------------------------------------------------------*/

eq.Dom.eqControl = function(el) {
   var c;
   while (el && el !== document && (c = el.classList) !== undefined) {
      if (c.contains('eq-control'))
         return el;
      if (c.contains('eq-container'))
         break;
      el = el.parentNode;
   }
   return null;
};

/*------------------------------------------------------------------------
   static Dom.eqOverlay()
   Obtain 'eq-overlay' element.
------------------------------------------------------------------------*/

eq.Dom.eqOverlay = function(el) {
   var c;
   while (el && el !== document && (c = el.classList) !== undefined) {
      if (c.contains('eq-overlay'))
         return el;
      el = el.parentNode;
   }
   return null;
};

/*------------------------------------------------------------------------
   static Dom.parentIndex()
   Return child element parent index.
------------------------------------------------------------------------*/

eq.Dom.parentIndex = function(el) {
   var i = 0;
   while (el = el.previousSibling)
      i++;
   return i;
};

/*------------------------------------------------------------------------
   static Dom.deleteChildren()
   Delete child elements.
------------------------------------------------------------------------*/

eq.Dom.deleteChildren = function(el) {
   var ch;
   while (ch = el.lastChild)
      el.removeChild(ch);
};

/*------------------------------------------------------------------------
   static Dom.inputElement()
   Return descendant input element.
------------------------------------------------------------------------*/

eq.Dom.inputElement = function(el) {
   var n = el.getElementsByTagName('INPUT');
   if (n.length !== 1)
      throw new Error("BUG: InputElement id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
};

/*------------------------------------------------------------------------
   static Dom.consumeEvent()
   Consume event.
------------------------------------------------------------------------*/

eq.Dom.consumeEvent = function(e) {
   e.preventDefault();
   e.stopPropagation();
};

/*------------------------------------------------------------------------
   static Dom.fireEvent()
   Fire event on element.
------------------------------------------------------------------------*/

eq.Dom.fireEvent = function(t, ty) {
   var e = document.createEvent("Event");
   e.initEvent(ty, true, true);
   t.dispatchEvent(e);
};

/*------------------------------------------------------------------------
   static Dom.fireMouseEvent()
   Fire mouse event on element.
------------------------------------------------------------------------*/

eq.Dom.fireMouseEvent = function(t, ty, ea) {
   var e = document.createEvent("MouseEvent");
   e.initEvent(ty, true, true);
   if (ea !== undefined)
      e.eq = ea;
   t.dispatchEvent(e);
};

/*------------------------------------------------------------------------
   static Dom.fireClick()
   Fire mouse click/double-click on element.
------------------------------------------------------------------------*/

eq.Dom.fireClick = function(t, ea) {
   this.fireMouseEvent(t, 'mousedown', ea);
   this.fireMouseEvent(t, 'mouseup', ea);
   this.fireMouseEvent(t, 'click', ea);
};

/*------------------------------------------------------------------------
   static Dom.isEnterOrSpace()
   Check for ENTER or SPACE key.
------------------------------------------------------------------------*/

eq.Dom.isEnterOrSpace = function(e) {
   switch (e.key) {
      case 'Enter':
         return 10;
      case ' ':
      case 'Spacebar': // IE, Firefox <= 37
         return 32;
   }
   return 0;
};

/*------------------------------------------------------------------------
   static Dom.onNextCycle()
   Invoke callback on next event cycle.
------------------------------------------------------------------------*/

eq.Dom.onNextCycle = function(cb, arg) {
   var self = this.onNextCycle, q = self.queue;
   q.push({ cb : cb, arg : arg });
   if (q.length === 1)
      window.setTimeout(self.exec);
};

eq.Dom.onNextCycle.queue = [];

eq.Dom.onNextCycle.exec = function() {
   var self = eq.Dom.onNextCycle, q = self.queue;
   for (var i = 0, l = q.length; i < l; i++) {
      var c = q[i], a = c.arg;
      if (a !== undefined)
         c.cb(a);
      else
         c.cb();
   }
   if (l) {
      q.splice(0, l);
      if (q.length)
         window.setTimeout(self.exec);
   }
};

/*------------------------------------------------------------------------
   static Dom.onNextFrame()
   Invoke callback on next animation frame.
------------------------------------------------------------------------*/

eq.Dom.onNextFrame = function(el, qi) {
   var self = this.onNextFrame, q = self.queue[self.queue_i];
   q.set(el, qi);
   if (q.size === 1)
      window.requestAnimationFrame(self.exec);
};

eq.Dom.onNextFrame.queue = [ new Map(), new Map() ];
eq.Dom.onNextFrame.queue_i = 0;

eq.Dom.onNextFrame.exec = function() {
   var self = eq.Dom.onNextFrame, q = self.queue[self.queue_i];
   if (++self.queue_i === self.queue.length)
      self.queue_i = 0;
   q.forEach(self.commit);
   q.clear();
};

eq.Dom.onNextFrame.commit = function(qi, el) {
   qi.cb(el, qi);
};

/*========================================================================
   DLG element callbacks.
========================================================================*/

eq.setAttr = new Map();
eq.FocusElement = new Map();
eq.SetFocus = new Map();
eq.ResetFocus = new Map();
eq.OnEvent = new Map();
eq.OnKey = new Map();
eq.OnChange = new Map();
eq.Changed = new Map();

/*------------------------------------------------------------------------
   CheckBox callbacks.
------------------------------------------------------------------------*/

eq.FocusElement.set('CheckBox', function(el) {
   return eq.Dom.inputElement(el);
});

eq.OnKey.set('CheckBox', function(id, el, ed, e) {
   var dm = eq.Dom, kc = dm.isEnterOrSpace(e);
   if (kc) {
      dm.fireClick(dm.inputElement(el), { key : kc });
      eq.c.App.debounceKey(e.key);
      return true;
   }
   return false;
});

eq.Changed.set('CheckBox', function(id, el) {
   var ch = {
      id : id,
      ty : eq.RsValType.buttonChecked,
      iv : el.checked ? 1 : 0
   };
   return ch;
});

/*------------------------------------------------------------------------
   ComboBox callbacks.
------------------------------------------------------------------------*/

eq.Dom.ComboBoxSet = function(rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   switch (rt.t) {
      case tag.items: {
         var
            v = rt.v,
            l = v.length,
            el = rt.el,
            c, n, nl, ln;
         if (!l)
            el.classList.add('eq-empty');
         if (!(c = el.querySelector('.eq-combobox>.eq-button'))) {
            c = document.createElement('DIV');
            c.className = 'eq-rt eq-button';
            el.appendChild(c);
         }
         if (!(c = el.querySelector('.eq-combobox>.eq-column'))) {
            c = document.createElement('DIV');
            c.className = 'eq-rt eq-column';
            el.appendChild(c);
         }
         n = c.childNodes;
         nl = n.length;
         for (var i = 0; i < l; i++) {
            if (i < nl)
               ln = n[i];
            else {
               ln = document.createElement('DIV');
               c.appendChild(ln);
            }
            dm.setContent(ln, v[i]);
         }
         while (n.length > l)
            c.removeChild(c.lastChild);
         break;
      }

      case tag.boxSize: {
         var
            c = rt.el.querySelector('.eq-combobox>.eq-column'),
            w = rt.w,
            h = rt.h,
            s = c.style;
         s.maxWidth = w ? w : '';
         s.maxHeight = h ? h : '';
         break;
      }

      default:
         throw new Error("ComboBoxSet: invalid tag " + rt.t);
   }
};

eq.setAttr.set('ComboBox', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.items: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.ComboBoxSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;
         }

         case tag.boxSize: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.ComboBoxSet, {
               el : el,
               t : rc.t,
               w : rc.w,
               h : rc.h
            });
            break;
         }

         default:
            throw new Error("setAttr.ComboBox: invalid tag " + rc.t);
      }
   }
});

eq.FocusElement.set('ComboBox', function(el) {
   return eq.Dom.inputElement(el);
});

eq.SetFocus.set('ComboBox', function(ct) {
   // Select on focus except on mouse click or if focus didn't change.
   if (eq.app.pendingMouseDown === undefined && ct !== document.activeElement)
      ct.select();
});

eq.ResetFocus.set('ComboBox', function(el) {
   if (el.classList.contains('eq-open'))
      eq.app.closeOverlay(el);
   var ct = el.childNodes[0];
   ct.selectionStart = ct.selectionEnd = 0;
   ct.blur();
});

eq.Dom.ComboBoxLine = function(t, co) {
   var c, cl;
   while (t) {
      if (!(c = t.parentNode))
         break;
      if (cl = c.classList) {
         if (cl.contains('eq-column')) {
            if (co !== undefined)
               co.c = c;
            return t;
         }
         if (cl.contains('eq-combobox')) {
            if (co !== undefined)
               co.ct = c;
            break;
         }
      }
      t = t.parentNode;
   }
   return undefined;
};

eq.Dom.ComboBoxSelect = function(el, c, ci) {
   if (c === undefined)
      if (!(c = el.querySelector('.eq-combobox>.eq-column')))
         c = document.body.querySelector('body>.eq-combobox.eq-column');
   // Deselect.
   var n = c.querySelectorAll('.eq-column>.eq-selected');
   for (var i = 0, l = n.length; i < l; i++)
      n[i].classList.remove('eq-selected');
   // Select.
   if (ci && c === ci.parentNode) {
      ci.classList.add('eq-selected');
      if ((n = c.childNodes).length > 1) {
         // Scroll into view if necessary.
         var
            bh = c.getBoundingClientRect().height,
            lh = n[0].getBoundingClientRect().height;
         if (bh > 0 && lh > 0) {
            var bt = c.scrollTop, lt = ci.offsetTop;
            if (lt < bt)
               c.scrollTop = lt;
            else if (lt + lh > bt + bh)
               c.scrollTop = Math.ceil(lt + lh - bh + 2);
         }
      }
   }
};

eq.Dom.ComboBoxOpen = function(el, c) {
   var cl = el.classList;
   if (!cl.contains('eq-open')) {
      if (c === undefined)
         c = el.querySelector('.eq-combobox>.eq-column');
      var
         dm = eq.Dom,
         v = el.childNodes[0].value,
         rc = el.getBoundingClientRect(),
         s = c.style,
         ccl = c.classList,
         ci, w;
      if (v.length) {
         // Pre-select current entry.
         var n = c.childNodes;
         for (var i = 0, l = n.length; i < l; i++)
            if (v === n[i].textContent) {
               ci = n[i];
               break;
            }
      }
      // Set overlay position and width.
      s.left = rc.left + window.pageXOffset + 'px';
      s.top = rc.top + rc.height + window.window.pageYOffset + 'px';
      if (w = s.maxWidth)
         s.width = w;
      else
         s.width = rc.width + 'px';
      // Make overlay.
      ccl.add('eq-combobox');
      if (cl.contains('eq-border'))
         ccl.add('eq-border');
      else
         ccl.remove('eq-border');
      eq.app.makeOverlay(el, c, dm.ComboBoxClose);
      cl.add('eq-open');
      cl.add('eq-enter');
      dm.ComboBoxSelect(el, c, ci);
   }
};

eq.Dom.ComboBoxClose = function(el, c, commit) {
   var cl = el.classList, ci;
   if (!cl.contains('eq-open'))
      throw new Error("ComboBoxClose BUG: not open, id:" + el.id);
   el.appendChild(c);
   c.classList.remove('eq-combobox');
   cl.remove('eq-open');
   cl.remove('eq-enter');
   if (commit && (ci = c.querySelector('.eq-column>.eq-selected'))) {
      var ct = el.childNodes[0];
      ct.value = ci.textContent;
      ct.select();
      eq.Dom.fireEvent(el, 'input');
   }
};

eq.OnKey.set('ComboBox', function(id, el, ed, e) {
   var
      dm = eq.Dom,
      c = el.querySelector('.eq-combobox>.eq-column'),
      d;
   switch (e.key) {
      case 'Enter':
         if (!c)
            eq.app.closeOverlay(el, true);
         return true;
      case 'Escape':
      case 'Esc': // IE
         if (!c)
            eq.app.closeOverlay(el, false);
         return true;

      // Navigation?
      case 'PageUp':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         if (c) {
            dm.ComboBoxOpen(el, c);
            d = 0;
         }
         break;
      default:
         // No.
         return false;
   }

   var n, ci;
   c = document.body.querySelector('body>.eq-combobox.eq-column');
   if (c && (n = c.childNodes).length) {
      if (d === undefined)
         switch (e.key) {
            case 'PageUp':
               ci = n[0];
               break;
            case 'PageDown':
               ci = n[n.length - 1];
               break;
            case 'ArrowUp':
            case 'Up':
               d = -1;
               break;
            case 'ArrowDown':
            case 'Down':
               d = 1;
               break;
         }
      if (d !== undefined) {
         if (ci = c.querySelector('.eq-column>.eq-selected')) {
            var i = dm.parentIndex(ci), l = n.length;
            if ((i += d) === -1)
               i = 0;
            else if (i >= l)
               i = l - 1;
            ci = n[i];
         }
         else
            ci = n[0];
      }
      if (ci !== undefined)
         dm.ComboBoxSelect(el, c, ci);
   }

   eq.app.lastKey = e.key;
   return true;
});

eq.Dom.ComboBoxMove = function(e, m) {
   var dm = eq.Dom, c = m.c;
   if (c) {
      var
         t = e.target,
         el = m.el,
         co = {},
         ln = dm.ComboBoxLine(t, co);
      if (el !== co.ct) {
         dm.ComboBoxSelect(el, c, ln);
         if (ln)
            m.close = true;
      }
   }
};

eq.Dom.ComboBoxMoved = function(m) {
   if (m.close)
      eq.app.closeOverlay(m.el, true);
   if (m.a !== undefined)
      m.a.classList.remove('eq-click');
};

eq.OnChange.set('ComboBox', function(id, el, e) {
   if (e.type === 'mousedown') {
      var app = eq.app, t = e.target;
      if (t && t.classList.contains('eq-button')) {
         if (app.setFocus(id, el)) {
            app.pendingFocus = {
               id : id,
               act : eq.Dom.ComboBoxOpen
            };
            return false;
         }
         var
            dm = eq.Dom,
            cl = el.classList,
            c = el.querySelector('.eq-combobox>.eq-column'),
            m = { el : el, a : t };
         if (cl.contains('eq-disabled') || cl.contains('eq-empty'))
            return false;
         t.classList.add('eq-click');
         if (c)
            dm.ComboBoxOpen(el, m.c = c);
         else
            m.close = true;
         app.moveCapture(m, dm.ComboBoxMove, dm.ComboBoxMoved);
         dm.consumeEvent(e);
         return false;
      }
      if (el.classList.contains('eq-open')) {
         var
            dm = eq.Dom,
            m = { el : el },
            co = {},
            ln = dm.ComboBoxLine(t, co);
         if (co.c) {
            dm.ComboBoxSelect(el, m.c = co.c, ln);
            m.close = true;
         }
         app.moveCapture(m, dm.ComboBoxMove, dm.ComboBoxMoved);
         if (el !== co.ct) {
            dm.consumeEvent(e);
            return false;
         }
      }
   }
   return true;
});

eq.Changed.set('ComboBox', function(id, el, ty) {
   for (var i = 0, l = ty.length; i < l; i++)
      if (ty[i] === 'input') {
         var ch = {
            id : id,
            ty : eq.RsValType.textValue,
            sv : el.value
         };
         return ch;
      }
   return null;
});

/*------------------------------------------------------------------------
   EditText callbacks.
------------------------------------------------------------------------*/

eq.FocusElement.set('EditText', function(el) {
   var n = el.getElementsByTagName('INPUT');
   if (n.length === 0)
      n = el.getElementsByTagName('TEXTAREA');
   if (n.length !== 1)
      throw new Error("BUG: Focus.EditText id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
});

eq.SetFocus.set('EditText', function(ct) {
   // Select on focus except on mouse click or if focus didn't change.
   if (   eq.app.pendingMouseDown === undefined
       && ct !== document.activeElement) {
      if (ct.tagName === 'INPUT')
         ct.select();
      else
         ct.selectionStart = ct.selectionEnd = 0;
   }
});

eq.ResetFocus.set('EditText', function(el) {
   var ct = el.childNodes[0];
   if (ct.tagName === 'INPUT')
      ct.selectionStart = ct.selectionEnd = 0;
   ct.blur();
});

eq.Changed.set('EditText', function(id, el, ty) {
   var v = [];
   for (var i = 0, l = ty.length; i < l; i++)
      switch (ty[i]) {
         case 'input':
            v.push({
               ty : eq.RsValType.textValue,
               sv : el.value
            });
            break;

         case 'blur': {
            var iv = el.selectionStart;
            if (iv >= 0)
               v.push({
                  ty : eq.RsValType.caretPos,
                  iv : iv
               });
            break;
         }
      }
   switch (v.length) {
      case 0:
         return null;
      case 1:
         v[0].id = id;
         return v[0];
      default:
         return {
            id : id,
            v : v
         };
   }
});

/*------------------------------------------------------------------------
   HtmlView callbacks.
------------------------------------------------------------------------*/

eq.Dom.HtmlView = {
   ready : function(e) {
      var
         dm = eq.Dom,
         self = dm.HtmlView,
         id = this.name,
         url = self.url.get(id),
         cn = self.content.get(id);
      if (url || cn)
         self.startTimeout();
      try {
         var
            w = this.contentWindow,
            d = w.document,
            h = d.head,
            b = d.body,
            n;
         w.name = id;
         if (cn) {
            b.style.margin = "0";
            dm.setHtmlContent(cn, b, h);
         }
         n = b.querySelectorAll('a[href^="eloq:"]');
         for (var i = 0, l = n.length; i < l; i++) {
            var a = n[i];
            a.dataset.eqLink = a.href;
            a.href = '#';
            a.addEventListener('click', self.clicked);
         }
      }
      catch (e) {
         console.log("HtmlView:" + id + (url ? " '" + url + "' " : " ") +
            e.name + ": " + e.message);
      };
   },

   clicked : function(e) {
      var
         id = e.view.name,
         el, ln;
      if (id && (el = document.getElementById(id))
             && (ln = this.dataset.eqLink)) {
         var
            self = eq.Dom.HtmlView,
            app = eq.app,
            r = self.re.exec(ln),
            rule;
         if (r) {
            if (r.length > 0 && r[1].length)
               rule = r[1];
            if (r.length > 2 && r[3])
               el.dataset.eqLink = r[3];
         }
         app.setChanged(id, 'change');
         if (rule !== undefined || el.classList.contains('eq-rule')) {
            if (rule !== undefined) {
               app.onceRule = rule;
               app.onceId = id;
            }
            app.submit(el, {
               id : id,
               type : 'change'
            });
         }
      }
      e.preventDefault();
      e.stopPropagation();
   },

   re : new RegExp('^eloq:([0-9]*)(/(.*))?$', 'i'),
   url : new Map(),
   content : new Map(),

   timeout : undefined,
   startTimeout : function() {
      if (this.timeout !== undefined)
         window.clearTimeout(this.timeout);
      this.timeout = window.setTimeout(this.timeoutElapsed, 1000);
   },
   timeoutElapsed : function() {
      var self = eq.Dom.HtmlView;
      self.timeout = undefined;
      self.url.clear();
      self.content.clear();
   }
};

eq.setAttr.set('HtmlView', function(dlg, id, el, rt) {
   var
      dm = eq.Dom,
      self = dm.HtmlView,
      tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.url:
            self.url.set(id, rc.v);
            self.content.delete(id);
            self.startTimeout();

            var ch = el.firstChild;
            if (ch && ch.tagName !== 'IFRAME') {
               el.removeChild(ch);
               ch = null;
            }
            if (!ch) {
               ch = document.createElement('IFRAME');
               ch.name = id;
               ch.onload = self.ready;
               el.appendChild(ch);
            }
            ch.src = rc.v;
            break;

         case tag.content: {
            var ch = el.firstChild;
            switch (rc.ty) {
               case 0: // text/html
                  self.url.delete(id);
                  self.content.set(id, rc.v);
                  self.startTimeout();

                  if (ch && ch.tagName !== 'IFRAME') {
                     el.removeChild(ch);
                     ch = null;
                  }
                  if (!ch) {
                     ch = document.createElement('IFRAME');
                     ch.name = id;
                     ch.onload = self.ready;
                     el.appendChild(ch);
                  }
                  ch.src = 'about:blank';
                  break;

               case 1: // text/plain
                  self.url.delete(id);
                  self.content.delete(id);
                  self.startTimeout();

                  if (ch && ch.tagName !== 'TEXTAREA') {
                     el.removeChild(ch);
                     ch = null;
                  }
                  if (!ch) {
                     ch = document.createElement('TEXTAREA');
                     ch.disabled = true;
                     ch.wrap = 'off'; // IE
                     el.appendChild(ch);
                  }
                  ch.value = ch.textContent = rc.v;
                  break;

               default:
                  throw new Error("setAttr.HtmlView: tag " + rc.t +
                     " invalid type " + rc.ty);
            }
            break;
         }

         default:
            throw new Error("setAttr.HtmlView: invalid tag " + rc.t);
      }
   }
});

eq.Changed.set('HtmlView', function(id, el) {
   var ln = el.dataset.eqLink;
   if (ln)
      delete el.dataset.eqLink;
   var ch = {
      id : id,
      ty : eq.RsValType.textValue,
      sv : ln ? ln : ""
   };
   return ch;
});

/*------------------------------------------------------------------------
   ListBox callbacks.
------------------------------------------------------------------------*/

eq.Dom.ListBoxSelect = function(el, li) {
   // Deselect.
   var n = el.querySelectorAll('.eq-column>.eq-selected');
   for (var i = 0, l = n.length; i < l; i++)
      n[i].classList.remove('eq-selected');
   // Select.
   if (li !== undefined) {
      n = el.querySelectorAll('.eq-column');
      for (var i = 0, l = n.length; i < l; i++) {
         var c = n[i].childNodes;
         if (li < c.length)
            c[li].classList.add('eq-selected');
      }
   }
};

eq.Dom.ListBoxSet = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.activeLine: {
         var v = rt.v;
         if (v > 0)
            eq.Dom.ListBoxSelect(rt.el, v - 1);
         else
            eq.Dom.ListBoxSelect(rt.el);
         break;
      }

      case tag.topItem: {
         var
            el = rt.el.querySelector('.eq-column'),
            n = el.childNodes,
            bx = el.parentNode,
            st = 0;
         if (n.length > 1) {
            var smax = el.getBoundingClientRect().height
                     - bx.getBoundingClientRect().height;
            if (smax > 0) {
               var lh = n[0].getBoundingClientRect().height;
               if (lh > 0 && (st = rt.v * lh) > smax)
                  st = Math.ceil(smax + 2);
            }
         }
         bx.scrollTop = st;
         break;
      }

      default:
         throw new Error("ListBoxSet: invalid tag " + rt.t);
   }
};

eq.Dom.ListBoxOnScroll = function(e) {
   if (eq.app.currDlg) {
      var dm = eq.Dom;
      dm.onNextFrame(e.target, { cb : dm.ListBoxOnScroll.exec });
   }
};

eq.Dom.ListBoxOnScroll.exec = function(el) {
   var cl = el.classList, lb;
   if (cl && cl.contains('eq-pane') && (lb = el.parentNode)
                                    && (cl = lb.classList)) {
      if (cl.contains('eq-listbox')) {
         // 'scroll' change event.
         eq.app.setChanged(lb.id, 'scroll');
         if (cl.contains('eq-multi')) {
            // Scroll multi-column ListBox header.
            var hc = el.previousSibling.firstChild;
            if (hc)
               hc.style.marginLeft = '-' + el.scrollLeft + 'px';
         }
      }
   }
};

eq.setAttr.set('ListBox', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.columns: {
            var
               cw = rc.cw,
               c;
            dm.deleteChildren(el);
            if (cw === undefined) {
               // Single-column ListBox.
               el.classList.remove('eq-multi');
               el.classList.add('eq-single');

               var p = document.createElement('DIV');
               p.className = 'eq-pane';
               p.addEventListener('scroll', dm.ListBoxOnScroll);

               c = document.createElement('DIV');
               c.className = 'eq-column';

               p.appendChild(c);
               el.appendChild(p);
            }
            else {
               // Multi-column ListBox.
               el.classList.remove('eq-single');
               el.classList.add('eq-multi');

               var
                  ct = rc.ct,
                  h = document.createElement('DIV'),
                  p = document.createElement('DIV');

               h.className = 'eq-caption';
               p.className = 'eq-pane';
               p.addEventListener('scroll', dm.ListBoxOnScroll);

               for (var i = 0, l = cw.length; i < l; i++) {
                  c = document.createElement('DIV');
                  c.style.width = cw[i];
                  dm.setContent(c, ct[i]);
                  h.appendChild(c);

                  c = document.createElement('DIV');
                  c.className = 'eq-column';
                  c.style.width = cw[i];
                  p.appendChild(c);
               }

               el.appendChild(h);
               el.appendChild(p);
            }
            break;
         }

         case tag.content: {
            var
               n = el.querySelectorAll('.eq-column'),
               nl = n.length;
            if (nl) {
               var
                  v = rc.v,
                  vl = v.length,
                  ci = rc.ci;
               if (ci === undefined) {
                  // Single-column ListBox.
                  var
                     c = n[0],
                     ln, s;
                  n = c.childNodes;
                  nl = n.length;
                  for (var vi = 0; vi < vl; vi++) {
                     if (vi < nl)
                        ln = n[vi];
                     else {
                        ln = document.createElement('DIV');
                        c.appendChild(ln);
                     }
                     if ((s = v[vi]).length)
                        dm.setContent(ln, s);
                     else
                        ln.textContent = ' ';
                  }
                  while (n.length > vl)
                     c.removeChild(c.lastChild);
               }
               else {
                  // Multi-column ListBox.
                  var
                     cs = rc.cs,
                     nv = [],
                     ln, s;
                  nv.length = nl;
                  for (var ni = 0; ni < nl; ni++)
                     nv[ni] = n[ni].childNodes;
                  for (var vi = 0; vi < vl; vi++) {
                     var
                        t = v[vi].split(cs),
                        tl = t.length;
                     for (var ni = 0; ni < nl; ni++) {
                        var
                           i = ci[ni],
                           nn = nv[ni];
                        if (vi < nn.length) {
                           ln = nn[vi];
                           if (i >= tl) {
                              ln.textContent = ' ';
                              continue;
                           }
                        }
                        else {
                           ln = document.createElement('DIV');
                           n[ni].appendChild(ln);
                        }
                        if (i < tl && (s = t[i]).length)
                           dm.setContent(ln, s);
                        else
                           ln.textContent = ' ';
                     }
                  }
                  for (var ni = 0; ni < nl; ni++) {
                     var
                        c = n[ni],
                        nn = nv[ni];
                     while (nn.length > vl)
                        c.removeChild(c.lastChild);
                  }
               }
            }
            break;
         }

         case tag.activeLine:
         case tag.topItem:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.ListBoxSet, {
               el : el,
               t  : rc.t,
               v  : rc.v
            });
            break;

         case tag.align: {
            var
               ca = rc.ca,
               l = ca.length,
               c = el.querySelectorAll('.eq-pane>.eq-column');
            if (l === c.length)
               for (var i = 0; i < l; i++) {
                  var st = c[i].style;
                  switch (ca[i]) {
                     case 0:
                        st.textAlign = '';
                        break;
                     case 1:
                        st.textAlign = 'center';
                        break;
                     case 2:
                        st.textAlign = 'right';
                  }
               }
            break;
         }

         default:
            throw new Error("setAttr.ListBox: invalid tag " + rc.t);
      }
   }
});

eq.OnKey.set('ListBox', function(id, el, ed, e) {
   var
      app = eq.app,
      dm = eq.Dom,
      kc = dm.isEnterOrSpace(e);
   if (kc) {
      var s = el.querySelector('.eq-column>.eq-selected');
      if (s) {
         e.eq = { key : kc };
         app.processEvent(e, id, el, ed);
      }
      eq.c.App.debounceKey(e.key);
      return true;
   }

   switch (e.key) {
      case 'Escape':
      case 'Esc': // IE
         dm.ListBoxSelect(el);
         eq.c.App.debounceKey(e.key);
         return true;

      // Navigation?
      case 'Home':
      case 'End':
      case 'PageUp':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         break;

      default:
         // No.
         return false;
   }

   var
      n = el.querySelector('.eq-listbox .eq-column').childNodes,
      d = 0,
      p = -1;
   if (n.length) {
      switch (e.key) {
         case 'Home':
            p = 0;
            kc = 36;
            break;
         case 'End':
            p = n.length - 1;
            kc = 35;
            break;
         case 'PageUp':
            d = -10;
            kc = 33;
            break;
         case 'PageDown':
            d = 10;
            kc = 34;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            kc = 38;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            kc = 40;
            break;
      }
      var ln = null, bx, bh, lh;
      if (p !== -1)
         ln = n[p];
      else if (d !== 0) {
         if ((d < -1 || d > 1) && n.length > 1) {
            // Page up/down, more than one line,
            // calculate number of visible lines.
            bx = n[0].parentNode.parentNode;
            bh = bx.getBoundingClientRect().height;
            lh = n[0].getBoundingClientRect().height;
            if (bh > 0 && lh > 0)
               d = d > 0 ? Math.floor(bh / lh) : Math.ceil(-(bh / lh));
         }
         var s = el.querySelector('.eq-column>.eq-selected');
         if (s) {
            p = dm.parentIndex(s);
            if ((p += d) < 0)
               p = 0;
            else if (p >= n.length)
               p = n.length - 1;

            ln = n[p];
         }
         else if (d > 0)
            ln = n[0];
      }

      if (ln !== null) {
         // Select line if necessary.
         if (!ln.classList.contains('eq-selected')) {
            app.setChanged(id, 'mousedown');
            dm.ListBoxSelect(el, dm.parentIndex(ln));
            if (n.length > 1) {
               // Scroll into view if necessary.
               if (bx === undefined) {
                  bx = n[0].parentNode.parentNode;
                  bh = bx.getBoundingClientRect().height;
                  lh = n[0].getBoundingClientRect().height;
               }
               if (bh > 0 && lh > 0) {
                  var bt = bx.scrollTop, lt = ln.offsetTop;
                  if (lt < bt)
                     bx.scrollTop = lt;
                  else if (lt + lh > bt + bh)
                     bx.scrollTop = Math.ceil(lt + lh - bh + 2);
               }
            }
         }
         // Submit on selection if enabled.
         if (ed.f) {
            e.eq = { key : kc };
            app.processEvent(e, id, el, ed);
         }
      }
   }

   app.lastKey = e.key;
   return true;
});

eq.OnEvent.set('ListBox', function(id, el, e) {
   if (e.type === 'keydown')
      return true;
   var t = e.target, cl, tp;
   while (t && (!(cl = t.classList) || !cl.contains('eq-listbox'))) {
      if (cl && cl.contains('eq-column'))
         break;
      if ((tp = t.parentNode) && (cl = tp.classList)
                              && cl.contains('eq-column'))
         return true;
      t = tp;
   }
   return false;
});

eq.OnChange.set('ListBox', function(id, el, e) {
   var dm = eq.Dom, t = e.target, cl, tp;
   while (t && (!(cl = t.classList) || !cl.contains('eq-listbox'))) {
      if (cl && cl.contains('eq-column')) {
         dm.ListBoxSelect(el);
         break;
      }
      if ((tp = t.parentNode) && (cl = tp.classList)
                              && cl.contains('eq-column')) {
         dm.ListBoxSelect(el, dm.parentIndex(t));
         break;
      }
      t = tp;
   }
   return true;
});

eq.Changed.set('ListBox', function(id, el, ty) {
   var v = [];
   for (var i = 0, l = ty.length; i < l; i++)
      switch (ty[i]) {
         case 'mousedown': {
            var ln = el.querySelector('.eq-selected');
            v.push({
               ty : eq.RsValType.activeLine,
               iv : ln ? eq.Dom.parentIndex(ln) + 1 : 0
            });
            break;
         }

         case 'scroll': {
            var
               n = el.querySelector('.eq-column').childNodes,
               iv = 0, lh;
            if (n.length > 1)
               if ((lh = n[0].getBoundingClientRect().height) > 0)
                  iv = Math.round(n[0].parentNode.parentNode.scrollTop / lh);
            v.push({
               ty : eq.RsValType.topItem,
               iv : iv
            });
            break;
         }
      }
   switch (v.length) {
      case 0:
         return null;
      case 1:
         v[0].id = id;
         return v[0];
      default:
         return {
            id : id,
            v : v
         };
   }
});

/*------------------------------------------------------------------------
   Menu callbacks.
------------------------------------------------------------------------*/

eq.Dom.isMenu = function(el) {
   var cl;
   while (el && el !== document) {
      if (cl = el.classList)
         if (cl.contains('eq-menubar') || cl.contains('eq-menu'))
            return true;
      el = el.parentNode;
   }
   return false;
};

eq.Dom.closeMenu = function() {
   var app = eq.app, id = app.menuOpen;
   if (id) {
      var el = document.getElementById(id);
      if (el && el.classList.contains('eq-open')) {
         el.classList.remove('eq-open');
         var n = el.getElementsByClassName('eq-open');
         for (var i = n.length; --i >= 0;)
            n[i].classList.remove('eq-open');
      }
      app.menuOpen = undefined;
   }
};

eq.Dom.MenuSet = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.text: {
         var
            el = rt.el,
            s = el.childNodes[0];
         if (!s || s.tagName !== 'SPAN') {
            var n = document.createElement('SPAN');
            s = el.insertBefore(n, s);
         }
         eq.Dom.setContent(s, rt.v, true);
         break;
      }

      default:
         throw new Error("MenuSet: invalid tag " + rt.t);
   }
};

eq.setAttr.set('Menu', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.text:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.MenuSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         case tag.align:
            if (rc.v)
               el.classList.add('eq-right');
            else
               el.classList.remove('eq-right');
            break;

         default:
            throw new Error("setAttr.Menu: invalid tag " + rc.t);
      }
   }
});

eq.OnChange.set('Menu', function(id, el, e) {
   var
      p = el.parentNode,
      cl,
      t = p && (cl = p.classList) && cl.contains('eq-menubar'),
      o = (cl = el.classList).contains('eq-open');
   if (t) {
      // MenuBar top-level menu.
      eq.Dom.closeMenu();
      if (!o) {
         cl.add('eq-open');
         eq.app.menuOpen = id;
      }
   }
   else if (o)
      cl.remove('eq-open');
   else {
      var mo = [ el ], tn;
      while (p) {
         if (   ((tn = (el = p).tagName) !== 'UL' || tn !== 'LI')
             || ((p = p.parentNode) && (cl = p.classList)
                                    && cl.contains('eq-menubar'))) {
            var n = el.getElementsByClassName('eq-open');
            for (var i = n.length; --i >= 0;)
               n[i].classList.remove('eq-open');
            break;
         }
         if ((cl = p.classList) && cl.contains('eq-menu'))
            mo.push(p);
      }
      for (var i = mo.length; --i >= 0;)
         mo[i].classList.add('eq-open');
   }
   return false;
});

eq.Changed.set('Menu', function() {
   return null;
});

/*------------------------------------------------------------------------
   MenuItem callbacks.
------------------------------------------------------------------------*/

eq.setAttr.set('MenuItem', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.text:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.MenuSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            throw new Error("setAttr.MenuItem: invalid tag " + rc.t);
      }
   }
});

eq.Changed.set('MenuItem', function(id, el) {
   return null;
});


/*------------------------------------------------------------------------
   PopText callbacks.
------------------------------------------------------------------------*/

eq.Dom.PopTextSet = function(rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   switch (rt.t) {
      case tag.content: {
         var
            v = rt.v,
            l = v.length,
            el = rt.el,
            c, n, nl, ln;
         if (!l)
            el.classList.add('eq-empty');
         if (!(c = el.querySelector('.eq-poptext>.eq-button'))) {
            c = document.createElement('DIV');
            c.className = 'eq-rt eq-button';
            el.appendChild(c);
         }
         if (!(c = el.querySelector('.eq-poptext>.eq-column'))) {
            c = document.createElement('DIV');
            c.className = 'eq-rt eq-column';
            el.appendChild(c);
         }
         n = c.childNodes;
         nl = n.length;
         for (var i = 0; i < l; i++) {
            if (i < nl)
               ln = n[i];
            else {
               ln = document.createElement('DIV');
               c.appendChild(ln);
            }
            dm.setContent(ln, v[i]);
         }
         while (n.length > l)
            c.removeChild(c.lastChild);
         break;
      }

      case tag.activeLine: {
         var
            el = rt.el,
            c = el.querySelector('.eq-poptext>.eq-column'),
            n = c.childNodes,
            v = rt.v,
            ci;
         if (v > 0 && v <= n.length)
            ci = n[v - 1];
         dm.copyContent(el.childNodes[0], ci);
         break;
      }

      case tag.boxSize: {
         var
            c = rt.el.querySelector('.eq-poptext>.eq-column'),
            w = rt.w,
            h = rt.h,
            s = c.style;
         s.maxWidth = w ? w : '';
         s.maxHeight = h ? h : '';
         break;
      }

      default:
         throw new Error("PopTextSet: invalid tag " + rt.t);
   }
};

eq.setAttr.set('PopText', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.content: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.PopTextSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;
         }

         case tag.activeLine: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.PopTextSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;
         }

         case tag.boxSize: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.PopTextSet, {
               el : el,
               t : rc.t,
               w : rc.w,
               h : rc.h
            });
            break;
         }

         default:
            throw new Error("setAttr.PopText: invalid tag " + rc.t);
      }
   }
});

eq.ResetFocus.set('PopText', function(el) {
   if (el.classList.contains('eq-open'))
      eq.app.closeOverlay(el);
});

eq.Dom.PopTextLine = function(t, co) {
   var c, cl;
   while (t) {
      if (!(c = t.parentNode))
         break;
      if (cl = c.classList) {
         if (cl.contains('eq-column')) {
            if (co !== undefined)
               co.c = c;
            return t;
         }
         if (cl.contains('eq-poptext')) {
            if (co !== undefined)
               co.ct = c;
            break;
         }
      }
      t = t.parentNode;
   }
   return undefined;
};

eq.Dom.PopTextSelect = function(el, c, ci) {
   if (c === undefined)
      if (!(c = el.querySelector('.eq-poptext>.eq-column')))
         c = document.body.querySelector('body>.eq-poptext.eq-column');
   // Deselect.
   var n = c.querySelectorAll('.eq-column>.eq-selected');
   for (var i = 0, l = n.length; i < l; i++)
      n[i].classList.remove('eq-selected');
   // Select.
   if (ci && c === ci.parentNode) {
      ci.classList.add('eq-selected');
      if ((n = c.childNodes).length > 1) {
         // Scroll into view if necessary.
         var
            bh = c.getBoundingClientRect().height,
            lh = n[0].getBoundingClientRect().height;
         if (bh > 0 && lh > 0) {
            var bt = c.scrollTop, lt = ci.offsetTop;
            if (lt < bt)
               c.scrollTop = lt;
            else if (lt + lh > bt + bh)
               c.scrollTop = Math.ceil(lt + lh - bh + 2);
         }
      }
   }
};

eq.Dom.PopTextOpen = function(el, c) {
   var cl = el.classList;
   if (!cl.contains('eq-open')) {
      if (c === undefined)
         c = el.querySelector('.eq-poptext>.eq-column');
      var
         dm = eq.Dom,
         v = el.childNodes[0].textContent,
         rc = el.getBoundingClientRect(),
         s = c.style,
         ccl = c.classList,
         ci, w;
      if (v.length) {
         // Pre-select current entry.
         var n = c.childNodes;
         for (var i = 0, l = n.length; i < l; i++)
            if (v === n[i].textContent) {
               ci = n[i];
               break;
            }
      }
      // Set overlay position and width.
      s.left = rc.left + window.pageXOffset + 'px';
      s.top = rc.top + rc.height + window.window.pageYOffset + 'px';
      if (w = s.maxWidth)
         s.width = w;
      else
         s.width = rc.width + 'px';
      // Make overlay.
      ccl.add('eq-poptext');
      if (cl.contains('eq-border'))
         ccl.add('eq-border');
      else
         ccl.remove('eq-border');
      eq.app.makeOverlay(el, c, dm.PopTextClose);
      cl.add('eq-open');
      cl.add('eq-enter');
      dm.PopTextSelect(el, c, ci);
   }
};

eq.Dom.PopTextClose = function(el, c, commit) {
   var cl = el.classList, ci;
   if (!cl.contains('eq-open'))
      throw new Error("PopTextClose BUG: not open, id:" + el.id);
   el.appendChild(c);
   c.classList.remove('eq-poptext');
   cl.remove('eq-open');
   cl.remove('eq-enter');
   if (commit && (ci = c.querySelector('.eq-column>.eq-selected'))) {
      var dm = eq.Dom;
      dm.copyContent(el.childNodes[0], ci);
      dm.fireEvent(el, 'input');
   }
};

eq.OnKey.set('PopText', function(id, el, ed, e) {
   var
      dm = eq.Dom,
      c = el.querySelector('.eq-poptext>.eq-column'),
      d;
   switch (e.key) {
      case 'Enter':
         if (!c)
            eq.app.closeOverlay(el, true);
         return true;
      case 'Escape':
      case 'Esc': // IE
         if (!c)
            eq.app.closeOverlay(el, false);
         return true;

      // Navigation?
      case 'PageUp':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         if (c) {
            dm.PopTextOpen(el, c);
            d = 0;
         }
         break;
      default:
         // No.
         return false;
   }

   var n, ci;
   c = document.body.querySelector('body>.eq-poptext.eq-column');
   if (c && (n = c.childNodes).length) {
      if (d === undefined)
         switch (e.key) {
            case 'PageUp':
               ci = n[0];
               break;
            case 'PageDown':
               ci = n[n.length - 1];
               break;
            case 'ArrowUp':
            case 'Up':
               d = -1;
               break;
            case 'ArrowDown':
            case 'Down':
               d = 1;
               break;
         }
      if (d !== undefined) {
         if (ci = c.querySelector('.eq-column>.eq-selected')) {
            var i = dm.parentIndex(ci), l = n.length;
            if ((i += d) === -1)
               i = 0;
            else if (i >= l)
               i = l - 1;
            ci = n[i];
         }
         else
            ci = n[0];
      }
      if (ci !== undefined)
         dm.PopTextSelect(el, c, ci);
   }

   eq.app.lastKey = e.key;
   return true;
});

eq.Dom.PopTextMove = function(e, m) {
   var dm = eq.Dom, c = m.c;
   if (c) {
      var
         t = e.target,
         el = m.el,
         co = {},
         ln = dm.PopTextLine(t, co);
      if (el !== co.ct) {
         eq.Dom.PopTextSelect(el, c, ln);
         if (ln)
            m.close = m.commit = true;
      }
   }
};

eq.Dom.PopTextMoved = function(m) {
   if (m.close)
      eq.app.closeOverlay(m.el, m.commit);
   if (m.a !== undefined)
      m.a.classList.remove('eq-click');
};

eq.OnChange.set('PopText', function(id, el, e) {
   if (e.type === 'mousedown') {
      var
         app = eq.app,
         dm = eq.Dom,
         t = e.target,
         ib = t && t.classList.contains('eq-button');
      if (!ib) {
         ib = el === dm.eqControl(t)
           && (t = el.querySelector('.eq-poptext>.eq-button'));
         if (el.classList.contains('eq-open')) {
            var
               m = { el : el },
               co = {},
               ln = dm.PopTextLine(t, co);
            if (co.c) {
               dm.PopTextSelect(el, m.c = co.c, ln);
               m.close = m.commit = true;
            }
            else if (el === co.ct) {
               m.a = t;
               m.close = true;
               t.classList.add('eq-click');
            }
            app.moveCapture(m, dm.PopTextMove, dm.PopTextMoved);
            if (el !== co.ct) {
               dm.consumeEvent(e);
               return false;
            }
            return true;
         }
      }

      if (ib) {
         if (app.setFocus(id, el)) {
            app.pendingFocus = {
               id : id,
               act : eq.Dom.PopTextOpen
            };
            return false;
         }
         var
            cl = el.classList,
            c = el.querySelector('.eq-poptext>.eq-column'),
            m = { el : el, a : t };
         if (cl.contains('eq-disabled') || cl.contains('eq-empty'))
            return false;
         t.classList.add('eq-click');
         if (c)
            dm.PopTextOpen(el, m.c = c);
         else
            m.close = true;
         app.moveCapture(m, dm.PopTextMove, dm.PopTextMoved);
         dm.consumeEvent(e);
         return false;
      }
   }
   return true;
});

eq.Changed.set('PopText', function(id, el, ty) {
   for (var i = 0, l = ty.length; i < l; i++)
      if (ty[i] === 'input') {
         var
            ci = el.querySelector('.eq-poptext>.eq-column>.eq-selected'),
            ch = {
               id : id,
               ty : eq.RsValType.activeLine,
               iv : ci ? eq.Dom.parentIndex(ci) + 1 : 0
            };
         return ch;
      }
   return null;
});

/*------------------------------------------------------------------------
   ProgressBar callbacks.
------------------------------------------------------------------------*/

eq.setAttr.set('ProgressBar', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.vertical: {
            if (!el.firstChild) {
               var
                  d = document.createElement('DIV'),
                  s1 = document.createElement('SPAN'),
                  s2 = document.createElement('SPAN');
               d.appendChild(s2);
               el.appendChild(s1);
               el.appendChild(d);
            }
            var cl = el.classList;
            if (rc.v) {
               cl.remove('eq-horz');
               cl.add('eq-vert');
            }
            else {
               cl.remove('eq-vert');
               cl.add('eq-horz');
            }
            break;
         }

         case tag.value: {
            var
               cl = el.classList,
               d = el.querySelector('.eq-progressbar>div'),
               v = rc.v;
            if (v) {
               cl.remove('eq-noval');
               cl.add('eq-val');
               if (cl.contains('eq-horz'))
                  d.style.width = v;
               else
                  d.style.height = v;
            }
            else {
               cl.remove('eq-val');
               cl.add('eq-noval');
               d.style.width = '';
            }
            break;
         }

         case tag.text: {
            var s = el.getElementsByTagName('SPAN');
            dm.setContent(s[0], rc.v);
            dm.setContent(s[1], rc.v);
            break;
         }

         default:
            throw new Error("setAttr.ProgressBar: invalid tag " + rc.t);
      }
   }
});

/*------------------------------------------------------------------------
   PushButton callbacks.
------------------------------------------------------------------------*/

eq.FocusElement.set('PushButton', function(el) {
   var n = el.getElementsByTagName('BUTTON');
   if (n.length !== 1)
      throw new Error("BUG: Focus.PushButton id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
});

eq.OnKey.set('PushButton', function(id, el, ed, e) {
   var dm = eq.Dom, kc = dm.isEnterOrSpace(e);
   if (kc) {
      dm.fireClick(el, { key : kc });
      eq.c.App.debounceKey(e.key);
      return true;
   }
   return false;
});

/*------------------------------------------------------------------------
   RadioButton callbacks.
------------------------------------------------------------------------*/

eq.FocusElement.set('RadioButton', function(el) {
   return eq.Dom.inputElement(el);
});

eq.OnEvent.set('RadioButton', function(id, el, e) {
   return !eq.Dom.inputElement(el).checked;
});

eq.OnKey.set('RadioButton', function(id, el, ed, e) {
   var dm = eq.Dom, kc = dm.isEnterOrSpace(e);
   if (kc) {
      dm.fireClick(dm.inputElement(el), { key : kc });
      eq.c.App.debounceKey(e.key);
      return true;
   }

   // Navigation?
   switch (e.key) {
      case 'Home':
      case 'PageUp':
      case 'End':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowLeft':
      case 'Left':
      case 'ArrowDown':
      case 'Down':
      case 'ArrowRight':
      case 'Right':
         break;
      default:
         // No.
         return false;
   }

   var
      t = e.target,
      n = el.parentNode.querySelectorAll('input[name="' + t.name + '"]'),
      d = 0;
   if (n.length) {
      switch (e.key) {
         case 'Home':
            el = n[0];
            kc = 36;
            break;
         case 'End':
            el = n[n.length - 1];
            kc = 35;
            break;
         case 'PageUp':
            el = n[0];
            kc = 33;
            break;
         case 'PageDown':
            el = n[n.length - 1];
            kc = 34;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            kc = 38;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            kc = 40;
            break;
         case 'ArrowLeft':
         case 'Left':
            d = -1;
            kc = 37;
            break;
         case 'ArrowRight':
         case 'Right':
            d = 1;
            kc = 39;
            break;
      }
      if (d !== 0) {
         el = null;
         for (var i = 0, l = n.length; i < l; i++)
            if (t === n[i]) {
               if ((i += d) === -1)
                  i = l - 1;
               else if (i >= l)
                  i = 0;
               el = n[i];
               break;
            }
      }
      if (el !== null)
         dm.fireClick(el, { key : kc });
   }

   eq.app.lastKey = e.key;
   return true;
});

eq.Changed.set('RadioButton', function(id, el) {
   if (!el.checked)
      return null;
   var ch = {
      id : id,
      ty : eq.RsValType.buttonChecked,
      iv : 1
   };
   return ch;
});

/*------------------------------------------------------------------------
   Splitter callbacks.
------------------------------------------------------------------------*/

eq.Dom.SplitterSet = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.panes: {
         var
            el = rt.el,
            n = el.childNodes;
         for (var i = 0, l = n.length; i < l; i++) {
            var
               c = n[i],
               cl = c.classList;
            if (cl.contains('eq-splitter') || cl.contains('eq-groupbox')) {
               var
                  pa = document.createElement('DIV'),
                  ca = document.createElement('DIV');
               el.insertBefore(pa, c);
               pa.appendChild(c);
               ca.classList.add('eq-caption');
               pa.appendChild(ca);
            }
         }
         break;
      }

      case tag.position: {
         var
            el = rt.el,
            n = el.childNodes;
         if (n.length === 2) {
            if (el.classList.contains('eq-vert'))
               n[0].style.height = rt.v + 'px';
            else
               n[0].style.width = rt.v + 'px';
         }
         break;
      }

      default:
         throw new Error("SplitterSet: invalid tag " + rt.t);
   }
};

eq.setAttr.set('Splitter', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.panes:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.SplitterSet, {
               el : el,
               t : rc.t
            });
            break;

         case tag.position:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.SplitterSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            throw new Error("setAttr.Splitter: invalid tag " + rc.t);
      }
   }
});

eq.Dom.SplitterMove = function(e, m) {
   if (!m.busy) {
      m.busy = true;
      var dm = eq.Dom;
      dm.onNextFrame(m.el, {
         cb : dm.SplitterMove.exec,
         x : e.clientX,
         y : e.clientY,
         m : m
      });
   }
};

eq.Dom.SplitterMove.exec = function(el, qi) {
   var
      x = qi.x,
      y = qi.y,
      m = qi.m,
      pa = m.pa,
      rc = pa.getBoundingClientRect();
   if (m.x !== undefined) {
      // Horizontal.
      var w = rc.width + x - m.x;
      if (w >= 0) {
         pa.style.width = w + 'px';
         if (pa.getBoundingClientRect().width !== rc.width) {
            m.x = x;
            m.modified = true;
         }
      }
   }
   else {
      // Vertical.
      var h = rc.height + y - m.y;
      if (h >= 0) {
         pa.style.height = h + 'px';
         if (pa.getBoundingClientRect().height !== rc.height) {
            m.y = y;
            m.modified = true;
         }
      }
   }

   m.busy = undefined;
};

eq.Dom.SplitterMoved = function(m) {
   eq.app.submit(m.el, {
      id : m.id,
      type : 'change'
   });
};

eq.OnChange.set('Splitter', function(id, el, e) {
   var t = e.target;
   if (t && t.classList.contains('eq-caption')) {
      var
         cl = el.classList,
         m = {
            id : id,
            el : el,
            pa : t.parentNode
         };
      if (cl.contains('eq-vert'))
         m.y = e.clientY;
      else
         m.x = e.clientX;
      eq.app.moveCapture(m,
         eq.Dom.SplitterMove,
         cl.contains('eq-rule') ? eq.Dom.SplitterMoved : null);
      e.stopPropagation();
   }
   return false;
});

eq.Changed.set('Splitter', function(id, el) {
   var n = el.childNodes;
   if (n.length === 2) {
      var
         rc = n[0].getBoundingClientRect(),
         ch = {
            id : id,
            ty : eq.RsValType.splitterPos,
            iv : el.classList.contains('eq-vert') ? rc.height : rc.width
         };
      return ch;
   }
   return null;
});

/*------------------------------------------------------------------------
   TabBox callbacks.
------------------------------------------------------------------------*/

eq.Dom.TabBoxSelect = function(el, ta) {
   var tp;
   if (ta !== undefined) {
      if ((tp = ta.parentNode) && tp.classList.contains('eq-tabs')) {
         if (ta.classList.contains('eq-selected')) {
            // Already selected.
            return null;
         }
      }
      else {
         // Not a tab.
         tp = undefined;
      }
   }
   // Deselect.
   var n = el.childNodes;
   for (var ni = 0, nl = n.length; ni < nl; ni++) {
      var tn = n[ni].childNodes;
      for (var i = 0, l = tn.length; i < l; i++)
         tn[i].classList.remove('eq-selected');
   }
   // Select.
   if (tp !== undefined) {
      var
         dm = eq.Dom,
         i = dm.parentIndex(ta),
         cl;
      ta.classList.add('eq-selected');
      if (tp = tp.nextSibling) {
         for (ta = tp.firstChild; ta; ta = ta.nextSibling) {
            if (   (cl = ta.classList)
                && (cl.contains('eq-hidden') || cl.contains('eq-disabled')))
               continue;
            if (i-- === 0) {
               ta.classList.add('eq-selected');
               return ta;
            }
         }
      }
   }
   return null;
};

eq.Dom.TabBoxSet = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.selectedTab: {
         var
            dm = eq.Dom,
            el = rt.el,
            v = rt.v,
            cl, tb, ta;
         if (v.length) {
            var n = el.childNodes[1].childNodes;
            for (var i = 0, j = 0, l = n.length; i < l; i++) {
               if (   (cl = (tb = n[i]).classList)
                   && (cl.contains('eq-hidden') || cl.contains('eq-disabled')))
                  continue;
               if (v === tb.id) {
                  ta = el.childNodes[0].childNodes[j];
                  break;
               }
               j++;
            }
         }
         dm.TabBoxSelect(el, ta);
         break;
      }

      default:
         throw new Error("TabBoxSet: invalid tag " + rt.t);
   }
};

eq.setAttr.set('TabBox', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.tabs: {
            var
               v = rc.v,
               tb = el.childNodes[0],
               ta, n, nl, s;
            if (!tb || !tb.classList.contains('eq-tabs')) {
               ta = document.createElement('DIV');
               ta.className = 'eq-rt eq-tabs';
               tb = el.insertBefore(ta, tb);
               dlg.updateControl(id, tb, {
                  ev : {
                     // Element is focusable and can submit,
                     // use 'mousedown', not 'click'.
                     c : [ 'mousedown' ]
                  }
               });
            }
            n = tb.childNodes;
            nl = n.length;
            for (var i = 0, l = v.length; i < l; i++) {
               if (i < nl)
                  ta = n[i];
               else {
                  ta = document.createElement('DIV');
                  tb.appendChild(ta);
               }
               if ((s = v[i]).length)
                  dm.setContent(ta, s, true);
               else
                  ta.textContent = ' ';
            }
            while (n.length > l)
               tb.removeChild(tb.lastChild);
            // Deselect, selectedTab always set afterwards.
            dm.TabBoxSelect(el);
            break;
         }

         case tag.selectedTab:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.TabBoxSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         case tag.rule:
            dlg.updateControl(id, el.childNodes[0], {
               ev : {
                  // Element is focusable and can submit,
                  // use 'mousedown', not 'click'.
                  a : rc.v ? [ 'mousedown' ] : null
               }
            });
            break;

         default:
            throw new Error("setAttr.TabBox: invalid tag " + rc.t);
      }
   }
});

eq.OnKey.set('TabBox', function(id, el, ed, e) {
   var dm = eq.Dom;

   // Navigation?
   switch (e.key) {
      case 'Home':
      case 'PageUp':
      case 'End':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowLeft':
      case 'Left':
      case 'ArrowDown':
      case 'Down':
      case 'ArrowRight':
      case 'Right':
         break;
      default:
         // No.
         return false;
   }

   var
      n = el.parentNode.querySelectorAll('.eq-tabs > *'),
      d = 0,
      kc;
   if (n.length) {
      switch (e.key) {
         case 'Home':
            el = n[0];
            kc = 36;
            break;
         case 'End':
            el = n[n.length - 1];
            kc = 35;
            break;
         case 'PageUp':
            el = n[0];
            kc = 33;
            break;
         case 'PageDown':
            el = n[n.length - 1];
            kc = 34;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            kc = 38;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            kc = 40;
            break;
         case 'ArrowLeft':
         case 'Left':
            d = -1;
            kc = 37;
            break;
         case 'ArrowRight':
         case 'Right':
            d = 1;
            kc = 39;
            break;
      }
      if (d !== 0) {
         el = null;
         for (var i = 0, l = n.length; i < l; i++)
            if (n[i].classList.contains('eq-selected')) {
               if ((i += d) === -1)
                  i = l - 1;
               else if (i >= l)
                  i = 0;
               el = n[i];
               break;
            }
      }
      if (el !== null)
         dm.fireClick(el, { key : kc });
   }

   eq.app.lastKey = e.key;
   return true;
});

eq.OnChange.set('TabBox', function(id, el, e) {
   if (e.type === 'mousedown') {
      var dm = eq.Dom, t = e.target, cl, tp, ta;
      while (t && (!(cl = t.classList) || !cl.contains('eq-tabbox'))) {
         if ((tp = t.parentNode) && (cl = tp.classList)
                                 && cl.contains('eq-tabs')) {
            if (ta = dm.TabBoxSelect(el, t)) {
               var app = eq.app;
               if (ta.classList.contains('eq-rule'))
                  app.submit(ta, {
                     id : ta.id,
                     type : 'change'
                  });
               else if (el.classList.contains('eq-rule'))
                  app.submit(el, {
                     id : id,
                     type : 'change'
                  });
               else
                  app.api.postMessage({
                     o : eq.ApiOp.updateTabOrder,
                     d : app.currDlg.id,
                     a : { id : id, arg : dm.parentIndex(t) }
                  });
            }
            break;
         }
         t = tp;
      }
   }
   return false;
});

eq.Changed.set('TabBox', function(id, el) {
   if (   el.classList.contains('eq-tabs')
       && (el = el.querySelector('.eq-tabs>.eq-selected'))) {
      var ch = {
         id : id,
         ty : eq.RsValType.selectedTab,
         iv : eq.Dom.parentIndex(el)
      };
      return ch;
   }

   return null;
});

/*------------------------------------------------------------------------
   Tree callbacks.
------------------------------------------------------------------------*/

eq.Dom.TreeSelect = function(el, i) {
   var ti, cl;
   if (i !== undefined) {
      var n = el.querySelectorAll('.eq-tree .eq-item');
      if (i >= 0 && i < n.length)
         cl = (ti = n[i]).classList;
   }
   if (ti) {
      // Expand parent nodes if necessary.
      var p = ti, pl;
      while ((p = p.parentNode) && (pl = p.classList)
                                && !pl.contains('eq-tree'))
         if (pl.contains('eq-node'))
            pl.add('eq-open');
      // Scroll into view if necessary.
      var
         bh = el.getBoundingClientRect().height,
         ih = ti.querySelector('.eq-item>span').getBoundingClientRect().height;
      if (bh > 0 && ih > 0) {
         var pt = el.scrollTop, it = ti.offsetTop;
         if (it < pt)
            el.scrollTop = Math.floor(it - ih / 2);
         else if (it + ih > pt + bh)
            el.scrollTop = Math.ceil(it + ih * 1.5 - bh);
      }
   }
   // Deselect.
   var n = el.querySelectorAll('.eq-tree .eq-selected');
   for (var ni = 0, nl = n.length; ni < nl; ni++)
      n[ni].classList.remove('eq-selected');
   // Select.
   if (ti) {
      cl.add('eq-selected');
      return true;
   }
   return false;
};

eq.Dom.TreeSelectedIndex = function(el, sl) {
   if (sl === undefined)
      sl = el.querySelector('.eq-tree .eq-selected');
   if (sl) {
      var n = el.querySelectorAll('.eq-tree .eq-item');
      for (var i = 0, l = n.length; i < l; i++)
         if (sl === n[i])
            return i;
   }
   return -1;
};

eq.Dom.TreeItemExpanded = function(ti) {
   var cl;
   while (ti && (!(cl = ti.classList) || !cl.contains('eq-tree'))) {
      if (cl && (cl.contains('eq-node') || cl.contains('eq-leaf'))) {
         while ((ti = ti.parentNode) && (cl = ti.classList)
                                     && !cl.contains('eq-tree')) {
            if (cl.contains('eq-node') && !cl.contains('eq-open'))
               return false;
         }
         return true;
      }
      ti = ti.parentNode;
   }
   return false;
};

eq.Dom.TreeSetContent = function(p, v, pn) {
   this.deleteChildren(p);
   var l = v.length;
   if (!l && pn !== undefined) {
      var cl = pn.classList;
      cl.add('eq-empty');
      cl.remove('eq-open');
   }
   for (var i = 0; i < l; i++) {
      var
         vv = v[i],
         ti = document.createElement('DIV'),
         sp = document.createElement('SPAN');
      if (typeof vv === 'object') {
         // Node.
         var
            bt = document.createElement('DIV'),
            ch = document.createElement('DIV');
         ti.className = 'eq-item eq-node';
         if (vv.f & 2)
            ti.classList.add('eq-open');
         bt.className = 'eq-button';
         ti.appendChild(bt);
         this.setContent(sp, vv.hc);
         ti.appendChild(sp);
         ch.className = 'eq-pane';
         this.TreeSetContent(ch, vv.ch, ti);
         ti.appendChild(ch);
      }
      else {
         // Leaf.
         ti.className = 'eq-item eq-leaf';
         this.setContent(sp, vv);
         ti.appendChild(sp);
      }
      p.appendChild(ti);
   }
};

eq.Dom.TreeSet = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.content:
         eq.Dom.TreeSetContent(rt.el.querySelector('.eq-tree>.eq-pane'), rt.v);
         break;

      case tag.activeLine: {
         var v = rt.v;
         if (v > 0)
            eq.Dom.TreeSelect(rt.el, v - 1);
         else
            eq.Dom.TreeSelect(rt.el);
         break;
      }

      case tag.status: {
         var
            el = rt.el,
            n = el.querySelectorAll('.eq-tree .eq-node'),
            v = rt.v,
            vi = 0,
            vl = v.length;
         for (var i = 0, l = n.length; i < l; i++) {
            var cl = n[i].classList;
            if (vi < vl && i === v[vi] && !cl.contains('eq-empty')) {
               cl.add('eq-open');
               vi++;
            }
            else
               cl.remove('eq-open');
         }
         break;
      }

      default:
         throw new Error("TreeSet: invalid tag " + rt.t);
   }
};

eq.setAttr.set('Tree', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.content:
         case tag.activeLine:
         case tag.status:
            // Depend on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.TreeSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            throw new Error("setAttr.Tree: invalid tag " + rc.t);
      }
   }
});

eq.OnEvent.set('Tree', function(id, el, e) {
   if (e.type === 'keydown')
      return true;
   var t = e.target, cl;
   while (t && (!(cl = t.classList) || !cl.contains('eq-tree'))) {
      if (t.tagName === 'SPAN') {
         while (   (t = t.parentNode)
                && (!(cl = t.classList) || !cl.contains('eq-tree'))) {
            if (cl && (cl.contains('eq-node') || cl.contains('eq-leaf')))
               return true;
         }
         return false;
      }
      t = t.parentNode;
   }
   return false;
});

eq.OnKey.set('Tree', function(id, el, ed, e) {
   var
      app = eq.app,
      dm = eq.Dom,
      kc = dm.isEnterOrSpace(e);
   if (kc) {
      var sl = el.querySelector('.eq-tree .eq-selected>span');
      if (sl) {
         e.eq = { key : kc };
         app.processEvent(e, id, el, ed);
      }
      eq.c.App.debounceKey(e.key);
      return true;
   }

   var d;
   switch (e.key) {
      case 'Escape':
      case 'Esc': // IE
         dm.TreeSelect(el);
         eq.c.App.debounceKey(e.key);
         return true;

      // Navigation?
      case 'Home':
      case 'PageUp':
      case 'End':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         break;

      // Collapse?
      case 'ArrowLeft':
      case 'Left':
         d = -1;
         break;

      // Expand?
      case 'ArrowRight':
      case 'Right':
         d = 1;
         break;
      default:
         // No.
         return false;
   }

   if (d !== undefined) {
      var sl = el.querySelector('.eq-tree .eq-node.eq-selected');
      if (sl) {
         var cl = sl.classList;
         if (d === 1) {
            if (!cl.contains('eq-open') && !cl.contains('eq-empty')) {
               cl.add('eq-open');
               app.setChanged(id, 'status');
            }
         }
         else {
            if (cl.contains('eq-open')) {
               cl.remove('eq-open');
               app.setChanged(id, 'status');
            }
         }
      }
      app.lastKey = e.key;
      return true;
   }

   var
      n = el.querySelectorAll('.eq-tree .eq-item'),
      l = n.length, i;
   if (l) {
      switch (e.key) {
         case 'Home':
            i = 0;
            kc = 36;
            break;
         case 'End':
            i = l - 1;
            kc = 35;
            break;
         case 'PageUp':
            i = 0;
            kc = 33;
            break;
         case 'PageDown':
            i = l - 1;
            kc = 34;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            kc = 38;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            kc = 40;
            break;
      }
      if (d !== undefined) {
         if ((i = dm.TreeSelectedIndex(el)) === -1)
            i = 0;
         else if ((i += d) === -1)
            i = 0;
         else if (i >= l)
            i = l - 1;
      }
      var j = i;
      while (!dm.TreeItemExpanded(n[j])) {
         if (d < 0) {
            if (--j === -1) {
               j = i;
               while (!dm.TreeItemExpanded(n[++j])) {}
               break;
            }
         }
         else {
            if (++j === l) {
               j = i;
               while (!dm.TreeItemExpanded(n[--j])) {}
               break;
            }
         }
      }

      if (dm.TreeSelect(el, j))
         app.setChanged(id, 'change');
      // Submit on selection if enabled.
      if (ed.f) {
         e.eq = { key : kc };
         app.processEvent(e, id, el, ed);
      }
   }

   app.lastKey = e.key;
   return true;
});

eq.Dom.TreeMoved = function(m) {
   m.ti.classList.remove('eq-click');
   eq.app.setChanged(m.id, 'status');
};

eq.OnChange.set('Tree', function(id, el, e) {
   if (e.type === 'mousedown') {
      var app = eq.app, t = e.target;
      if (t) {
         if (app.setFocus(id, el))
            return;
         if (t.classList.contains('eq-button')) {
            var ti = t.parentNode, cl;
            if (ti && (cl = ti.classList) && cl.contains('eq-node')) {
               var ck = true;
               if (cl.contains('eq-open'))
                  cl.remove('eq-open');
               else if (!cl.contains('eq-empty'))
                  cl.add('eq-open');
               else
                  ck = false;
               if (ck) {
                  cl.add('eq-click');
                  app.moveCapture({ id : id, ti : ti }, null, eq.Dom.TreeMoved);
               }
            }
         }
         else {
            var cl;
            while (t && (!(cl = t.classList) || !cl.contains('eq-tree'))) {
               if (cl && cl.contains('eq-item')) {
                  var dm = eq.Dom, i = dm.TreeSelectedIndex(el, t);
                  if (dm.TreeSelect(el, i)) {
                     app.setChanged(id, 'change');
                     return true;
                  }
                  break;
               }
               t = t.parentNode;
            }
         }
      }
   }
   return false;
});

eq.Changed.set('Tree', function(id, el, ty) {
   var v = [];
   for (var i = 0, l = ty.length; i < l; i++)
      switch (ty[i]) {
         case 'change':
            v.push({
               ty : eq.RsValType.activeLine,
               iv : eq.Dom.TreeSelectedIndex(el) + 1
            });
            break;

         case 'status': {
            var
               n = el.querySelectorAll('.eq-tree .eq-node'),
               vv = [];
            for (var ni = 0, nl = n.length; ni < nl; ni++)
               if (n[ni].classList.contains('eq-open'))
                  vv.push(ni);
            v.push({
               ty : eq.RsValType.status,
               vv : vv
            });
            break;
         }
      }
   switch (v.length) {
      case 0:
         return null;
      case 1:
         v[0].id = id;
         return v[0];
      default:
         return {
            id : id,
            v : v
         };
   }
});

/*------------------------------------------------------------------------
   WindowMenu callbacks.
------------------------------------------------------------------------*/

eq.setAttr.set('WindowMenu', function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.text:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.MenuSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            throw new Error("setAttr.WindowMenu: invalid tag " + rc.t);
      }
   }
});

eq.OnChange.set('WindowMenu', function(id, el, e) {
   var
      dm = eq.Dom,
      o = el.classList.contains('eq-open');
   dm.closeMenu();
   if (o) {
      // WindowMenu clicked when open.
      var t = e.target;
      while ((t = t.parentNode) && t !== el)
         if (t.classList.contains('eq-item')) {
            eq.c.App.activateSession(eq.app.sessionByIndex(dm.parentIndex(t)));
            break;
         }
   }
   else {
      // Open WindowMenu.
      var
         app = eq.app,
         ml = el.lastChild,
         tl = app.sessionTitles();
      dm.deleteChildren(ml);
      for (var i = 0, l = tl.length; i < l; i++) {
         var
            mi = document.createElement('LI'),
            mt = document.createElement('SPAN');
         mi.className = 'eq-item';
         dm.setContent(mt, tl[i]);
         mi.appendChild(mt);
         ml.appendChild(mi);
      }
      el.classList.add('eq-open');
      app.menuOpen = id;
   }
   return true;
});

eq.Changed.set('WindowMenu', function() {
   return null;
});

/** @preserve $Id: 05-window.js,v 28.2 2021/09/23 13:30:39 mike Exp $
*//*======================================================================
   class Window extends Obj
   Window class.
========================================================================*/

eq.c.Obj.subClass('Window', function(root, parent, dlg) {
   eq.c.Obj.call(this);

   if (   !root.classList.contains('eq-window')
       && !root.classList.contains('eq-dialog'))
      throw new Error("BUG: Invalid window root");

   // Add to app window list.

   var app = eq.app, wl, wr, i, l, s = root.style;
   if ((wl = app.wl) === undefined)
      app.wl = [];
   else if (wl.length === 0)
      wl = undefined;

   if (!parent)
      parent = document.body;

   if (wl === undefined) {
      wl = app.wl;
      s.zIndex = eq.Dom.zIndex(parent) + 1;
   }
   else {
      for (i = 0, l = wl.length;;) {
         if ((wr = wl[i].root) === undefined)
            throw new Error("BUG: Window root undefined");
         if (++i === l) {
            wr.classList.remove('eq-active');
            s.zIndex = eq.Dom.zIndex(wr) + 1;
            break;
         }
      }
   }

   this.root = root;
   this.dlg = dlg;
   this.modified = {};
   wl.push(this);
   s.visibility = 'hidden';
   parent.appendChild(root);
});

/*------------------------------------------------------------------------
   Window.close()
   Close Window.
------------------------------------------------------------------------*/

eq.c.Window.prototype.close = function() {
   var r, wl, i, l, z, wr;
   if ((r = this.root) === undefined)
      throw new Error("BUG: Window root undefined");

   // Remove from app window list.

   if ((wl = eq.app.wl) === undefined)
      throw new Error("BUG: Window list undefined");

   for (i = wl.length; --i >= 0;)
      if (wl[i] === this) {
         wl.splice(i, 1);
         z = eq.Dom.zIndex(r);
         for (l = wl.length; i < l; i++) {
            if ((wr = wl[i].root) === undefined)
               throw new Error("BUG: Window root undefined");
            wr.style.zIndex = z++;
         }
         break;
      }
   if (i === -1)
      throw new Error("BUG: Window not in list");

   r.parentNode.removeChild(r);
   this.root = undefined;
   this.dlg = undefined;
   this.modified = undefined;
   this.pos = undefined;
   this.whenVisible = undefined;
   this.forcedInline = undefined;

   if (l = wl.length)
      return wl[l - 1];

   return null;
};

/*------------------------------------------------------------------------
   static Window.setupFrame()
   Setup Window frame.
------------------------------------------------------------------------*/

eq.c.Window.setupFrame = function(child) {
   if (   !child.classList.contains('eq-dialog')
       && !child.classList.contains('eq-popup-box'))
      throw new Error("BUG: Invalid window frame child");
   var
      f = document.createElement('DIV'),
      p = document.createElement('DIV'),
      c = document.createElement('DIV'),
      l = document.createElement('DIV'),
      s = document.createElement('SPAN'),
      m = document.createElement('DIV'),
      x = document.createElement('DIV'),
      i, n;
   f.className = 'eq-window';
   p.className = 'eq-pane';
   c.className = 'eq-caption';
   l.className = 'eq-logo eq-move';
   c.appendChild(l);
   s.className = 'eq-move';
   c.appendChild(s);
   m.className = 'eq-max';
   c.appendChild(m);
   x.className = 'eq-close';
   c.appendChild(x);
   p.appendChild(c);
   p.appendChild(child);
   f.appendChild(p);
   for (i = 0; i < 8; i++) {
      switch (i) {
         case 0:
            n = 'eq-top';
            break;
         case 1:
            n = 'eq-top eq-right';
            break;
         case 2:
            n = 'eq-right';
            break;
         case 3:
            n = 'eq-bottom eq-right';
            break;
         case 4:
            n = 'eq-bottom';
            break;
         case 5:
            n = 'eq-bottom eq-left';
            break;
         case 6:
            n = 'eq-left';
            break;
         case 7:
            n = 'eq-top eq-left';
      }
      p = document.createElement('DIV');
      p.className = n + ' eq-resize';
      f.appendChild(p);
   }
   return f;
};

/*------------------------------------------------------------------------
   Window.setPos()
   Set Window position.
------------------------------------------------------------------------*/

eq.c.Window.prototype.setPos = function(x, y) {
   if (x === -1 || y === -1)
      eq.Dom.onNextFrame(this, {
         cb : eq.c.Window.setPos.center,
         x : x,
         y : y
      });
   else {
      var r;
      if ((r = this.root) === undefined)
         throw new Error("BUG: Window root undefined");
      eq.c.Window.setPos(this, r, x, y);
   }
};

eq.c.Window.setPos = function(w, r, x, y) {
   var s = r.style, v;
   if (+x === +x)
      s.left = x > 0 ? x + 'px' : 0;
   else
      s.left = x;
   if (+y === +y)
      s.top = y > 0 ? y + 'px' : 0;
   else
      s.top = y;

   s.visibility = '';
   if ((v = w.whenVisible) !== undefined) {
      w.whenVisible = undefined;
      eq.Dom.onNextFrame(w, v);
   }
};

eq.c.Window.setPos.center = function(w, qi) {
   var r, rrc, x, y, wl, i, p, prc, pw, pwrc;
   if ((r = w.root) === undefined)
      throw new Error("BUG: Window root undefined");
   rrc = r.getBoundingClientRect();
   x = qi.x;
   y = qi.y;
   if ((wl = eq.app.wl) !== undefined)
      for (i = wl.length; --i >= 0;)
         if (wl[i] === w) {
            if (i > 0) {
               if ((pw = wl[i - 1].root) === undefined)
                  throw new Error("BUG: Window root undefined");
               if ((p = r.offsetParent) !== document.body)
                  prc = p.getBoundingClientRect();
               pwrc = pw.getBoundingClientRect();
               if (x === -1) {
                  x = (pwrc.width - rrc.width) / 2 + pwrc.left;
                  if (prc !== undefined)
                     x -= prc.left;
               }
               if (y === -1) {
                  y = (pwrc.height - rrc.height) / 2 + pwrc.top;
                  if (prc !== undefined)
                     y -= prc.top;
               }
               eq.c.Window.setPos(w, r, x, y);
               return;
            }
            break;
         }

   if ((p = r.offsetParent) === document.body) {
      var v = eq.Dom.viewRect();
      if (x === -1)
         x = (v.w - rrc.width) / 2 + v.x;
      if (y === -1)
         y = (v.h - rrc.height) / 2 + v.y;
   }
   else {
      prc = p.getBoundingClientRect();
      if (x === -1)
         x = (prc.width - rrc.width) / 2;
      if (y === -1)
         y = (prc.height - rrc.height) / 2;
   }

   eq.c.Window.setPos(w, r, x, y);
};

/*------------------------------------------------------------------------
   Window.setTitle()
   Set Window title.
------------------------------------------------------------------------*/

eq.c.Window.prototype.setTitle = function(t) {
   var r, c;
   if ((r = this.root) === undefined)
      throw new Error("BUG: Window root undefined");
   if (c = r.querySelector('.eq-window>.eq-pane>.eq-caption>span'))
      c.textContent = t;
};

/*------------------------------------------------------------------------
   Window.activate()
   Activate Window.
------------------------------------------------------------------------*/

eq.c.Window.prototype.activate = function() {
   var r, wl, i, l, z, wr, cl;
   if ((r = this.root) === undefined)
      throw new Error("BUG: Window root undefined");

   if ((wl = eq.app.wl) === undefined)
      throw new Error("BUG: Window list undefined");

   // Make topmost in app window list.

   for (l = wl.length, i = l; --i >= 0;) {
      if (wl[i] === this) {
         if (i !== l - 1) {
            wl.splice(i, 1);
            z = eq.Dom.zIndex(r);
            for (l = wl.length; i < l;) {
               if ((wr = wl[i].root) === undefined)
                  throw new Error("BUG: Window root undefined");
               wr.style.zIndex = z++;
               if (++i === l)
                  wr.classList.remove('eq-active');
            }
            r.style.zIndex = z;
            wl.push(this);
         }
         r.classList.add('eq-active');
         break;
      }
   }
   if (i === -1)
      throw new Error("BUG: Window not in list");
};

/*------------------------------------------------------------------------
   Window.makeInline()
   Convert Window to inline Dialog if necessary.
------------------------------------------------------------------------*/

eq.c.Window.prototype.makeInline = function(save) {
   var r, d, p, c, s;
   if ((r = this.root) === undefined)
      throw new Error("BUG: Window root undefined");
   if (r.classList.contains('eq-window')) {
      if (!(d = r.querySelector('.eq-window>.eq-pane>.eq-dialog')))
         throw new Error("BUG: no Dialog in Window");
      if (save) {
         p = d.querySelector('.eq-dialog>.eq-pane');
         c = r.querySelector('.eq-window>.eq-pane>.eq-caption>span');
         this.forcedInline = {
            x : r.offsetLeft,
            y : r.offsetTop,
            w : p.offsetWidth,
            h : p.offsetHeight,
            t : c ? c.textContent : undefined,
            r : r.classList.contains('eq-resize'),
            c : r.classList.contains('eq-close'),
            m : r.classList.contains('eq-max')
         };
      }
      (s = d.style).visibility = 'hidden';
      s.zIndex = eq.Dom.zIndex(r);
      d.classList.add('eq-inline');
      r.parentNode.replaceChild(d, r);
      return this.root = d;
   }
   return null;
};

/*------------------------------------------------------------------------
   Window.makeFramed()
   Convert inline Dialog to Window if necessary.
------------------------------------------------------------------------*/

eq.c.Window.prototype.makeFramed = function(restore) {
   var d, fi = this.forcedInline, parent, z, r, s, cl, p;
   if ((d = this.root) === undefined)
      throw new Error("BUG: Window root undefined");
   if (d.classList.contains('eq-dialog') && (!restore || fi !== undefined)) {
      z = eq.Dom.zIndex(d);
      r = eq.c.Window.setupFrame((parent = d.parentNode).removeChild(d));
      (s = r.style).visibility = 'hidden';
      s.zIndex = z;
      d.style.zIndex = '';
      (cl = d.classList).remove('eq-inline');
      cl.remove('eq-max');
      if (restore) {
         this.forcedInline = undefined;
         p = d.querySelector('.eq-dialog>.eq-pane');
         s.left = fi.x > 0 ? fi.x + 'px' : 0;
         s.top = fi.y > 0 ? fi.y + 'px' : 0;
         (s = p.style).width = fi.w > 0 ? fi.w + 'px' : 0;
         s.height = fi.h > 0 ? fi.h + 'px' : 0;
         cl = r.classList;
         if (fi.t !== undefined)
            r.querySelector(
               '.eq-window>.eq-pane>.eq-caption>span').textContent = fi.t;
         if (fi.r)
            cl.add('eq-resize');
         else
            cl.remove('eq-resize');
         if (fi.c)
            cl.add('eq-close');
         else
            cl.remove('eq-close');
         if (fi.m)
            cl.add('eq-max');
         else
            cl.remove('eq-max');
      }
      parent.appendChild(r);
      return this.root = r;
   }
   return null;
};

/*------------------------------------------------------------------------
   static Window.zOrderModified()
   Window list stack order has been modified.
------------------------------------------------------------------------*/

eq.c.Window.zOrderModified = function() {
   var app = eq.app, wl, lm, i, w, r, active, hide;
   if ((wl = app.wl) !== undefined) {

      // Inline base Dialog, forced by layout mode?
      if (   (lm = app.layoutMode) !== undefined
          && (!(lm & eq.layoutMode.inline) || !(lm & eq.layoutMode.dialog)))
         lm = undefined;

      // Adjust Windows from top to bottom.

      for (active = true, i = wl.length; --i >= 0;) {
         w = wl[i];
         if ((r = w.root) === undefined)
            throw new Error("BUG: Window root undefined");
         if (lm !== undefined && w.dlg !== undefined) {
            if (i === 0) {
               // Base Dialog, make inline.
               if (r = w.makeInline(true)) {
                  if (lm & eq.layoutMode.maximized)
                     r.classList.add('eq-max');
                  else
                     r.classList.remove('eq-max');
                  r.style.visibility = '';
               }
            }
            else {
               // Make framed if previously made inline.
               if (r = w.makeFramed(true))
                  r.style.visibility = '';
            }
         }
         if ((r = w.root) === undefined)
            throw new Error("BUG: Window root undefined");
         cl = r.classList;
         if (active === undefined)
            cl.remove('eq-active');
         else {
            active = undefined;
            cl.add('eq-active');
         }
         if (hide)
            cl.add('eq-hidden');
         else {
            cl.remove('eq-hidden');
            // Windows below topmost inline Dialog become invisible.
            if (!cl.contains('eq-window'))
               hide = true;
         }
      }
   }
};

/*------------------------------------------------------------------------
   static Window.fromTarget()
   Obtain Window from target element.
------------------------------------------------------------------------*/

eq.c.Window.fromTarget = function(t) {
   var wl, cl, i, w;

   // Locate target in app window list.

   if ((wl = eq.app.wl) !== undefined) {
      while (t && t !== document) {
         if (cl = t.classList) {
            if (cl.contains('eq-dialog') || cl.contains('eq-popup-box'))
               break;
            if (cl.contains('eq-window')) {
               for (i = wl.length; --i >= 0;)
                  if ((w = wl[i]).root === t)
                     return w;
               break;
            }
         }
         t = t.parentNode;
      }
   }
   return null;
};

/*------------------------------------------------------------------------
   Window.onMouseDownOrClick()
   Window 'mousedown' and 'click' event handler.
------------------------------------------------------------------------*/

eq.c.Window.prototype.onMouseDownOrClick = function(e) {
   if (e.type === 'mousedown') {
      var
         r = this.root,
         t = e.target,
         p = t.parentNode,
         cl = t.classList,
         app, mode;
      if (p && cl) {
         app = eq.app;
         if (cl.contains('eq-resize') && p === r) {
            if (cl.contains('eq-top')) {
               mode = 18;
               if (cl.contains('eq-left'))
                  mode += 1;
               else if (cl.contains('eq-right'))
                  mode += 4;
            }
            else if (cl.contains('eq-bottom')) {
               mode = 8;
               if (cl.contains('eq-left'))
                  mode += 17;
               else if (cl.contains('eq-right'))
                  mode += 4;
            }
            else if (cl.contains('eq-left'))
               mode = 17;
            else if (cl.contains('eq-right'))
               mode = 4;
         }
         else if (   p.classList.contains('eq-caption')
                  && (p = p.parentNode) && p.classList.contains('eq-pane')
                  && (p = p.parentNode) && p === r) {
            if (cl.contains('eq-close')) {
               cl.add('eq-click');
               var d;
               if ((d = this.dlg) !== undefined) {
                  if (d === app.currDlg)
                     app.submit(document.body, { type : 'close' });
                  else
                     d.pendingClose = true;
               }
               return true;
            }
            if (cl.contains('eq-max')) {
               if ((cl = p.classList).contains('eq-max')) {
                  cl.remove('eq-max');
                  var pos = this.pos;
                  if (pos !== undefined) {
                     this.pos = undefined;
                     eq.c.Window.setPos(this, r, pos.x, pos.y);
                  }
               }
               else {
                  this.pos = { x : r.offsetLeft, y : r.offsetTop };
                  cl.add('eq-max');
               }
               this.modified.maximized = true;
               return true;
            }
            if (cl.contains('eq-move'))
               if (!p.classList.contains('eq-max'))
                  mode = 3;
         }
         if (mode !== undefined) {
            var m = {
               w : this,
               x : e.clientX,
               y : e.clientY,
               mode : mode
            };
            if ((p = r.offsetParent) === document.body)
               m.px = m.py = 0;
            else {
               var prc = p.getBoundingClientRect();
               m.px = prc.left;
               m.py = prc.top;
            }
            app.moveCapture(m, eq.c.Window.onMove);
            return true;
         }
      }
   }
   return false;
};

/*------------------------------------------------------------------------
   static Window.onMove()
   Move Window.
------------------------------------------------------------------------*/

eq.c.Window.onMove = function(e, m) {
   if (!m.busy) {
      m.busy = true;
      eq.Dom.onNextFrame(m.w, {
         cb : eq.c.Window.onMove.exec,
         x : e.clientX,
         y : e.clientY,
         m : m
      });
   }
};

eq.c.Window.onMove.exec = function(w, qi) {
   var
      cx = qi.x,
      cy = qi.y,
      m = qi.m,
      mode = m.mode,
      sz = mode & 16,
      r = w.root,
      rc, chk;
   if (r === undefined)
      throw new Error("BUG: Window root undefined");
   rc = r.getBoundingClientRect();
   if (mode & 1) {
      // Left position modified.
      if (cx !== m.x && (m.maxX === undefined || cx <= m.maxX)) {
         var
            left = rc.left - m.px,
            x = left + cx - m.x;
         if (x < 0) {
            x = 0;
            cx = m.x - left;
         }
         r.style.left = x ? x + 'px' : 0;
         w.modified.pos = true;
         if (!sz)
            m.x = cx;
         else {
            // Width modified.
            if (chk === undefined) {
               var p = r.querySelector('.eq-dialog>.eq-pane');
               chk = { rc : rc, p : p, prc : p.getBoundingClientRect() };
            }
            chk.p.style.width = (left + chk.prc.width - x) + 'px';
            chk.cx = cx;
            chk.cw = rc.width;
         }
      }
   }
   if (mode & 2) {
      // Top position modified.
      if (cy !== m.y && (m.maxY === undefined || cy <= m.maxY)) {
         var
            top = rc.top - m.py,
            y = top + cy - m.y;
         if (y < 0) {
            y = 0;
            cy = m.y - top;
         }
         r.style.top = y ? y + 'px' : 0;
         w.modified.pos = true;
         if (!sz)
            m.y = cy;
         else {
            // Height modified.
            if (chk === undefined) {
               var p = r.querySelector('.eq-dialog>.eq-pane');
               chk = { rc : rc, p : p, prc : p.getBoundingClientRect() };
            }
            chk.p.style.height = (top + chk.prc.height - y) + 'px';
            chk.cy = cy;
            chk.cw = rc.height;
         }
      }
   }
   if (mode & 4) {
      // Width modified.
      if (cx !== m.x && (m.minX === undefined || cx >= m.minX)) {
         if (chk === undefined) {
            var p = r.querySelector('.eq-dialog>.eq-pane');
            chk = { rc : rc, p : p, prc : p.getBoundingClientRect() };
         }
         chk.p.style.width = (chk.prc.width + cx - m.x) + 'px';
         chk.cx = cx;
         chk.cw = rc.width;
      }
   }
   if (mode & 8) {
      if (cy !== m.y && (m.minY === undefined || cy >= m.minY)) {
         // Height modified.
         if (chk === undefined) {
            var p = r.querySelector('.eq-dialog>.eq-pane');
            chk = { rc : rc, p : p, prc : p.getBoundingClientRect() };
         }
         chk.p.style.height = (chk.prc.height + cy - m.y) + 'px';
         chk.cy = cy;
         chk.ch = rc.height;
      }
   }
   if (chk === undefined)
      m.busy = undefined;
   else
      eq.Dom.onNextFrame(w, {
         cb : eq.c.Window.onMove.chk,
         m : m,
         chk : chk
      });
};

eq.c.Window.onMove.chk = function(w, qi) {
   var
      m = qi.m,
      r = w.root,
      chk = qi.chk,
      rc = r.getBoundingClientRect();
   // New Window size may need to stabilize, could require multiple frames.
   if (   (chk.cw !== undefined && chk.cw !== rc.width)
       || (chk.ch !== undefined && chk.ch !== rc.height)) {
      if (chk.cw !== undefined)
         chk.cw = rc.width;
      if (chk.ch !== undefined)
         chk.ch = rc.height;
   }
   // When done, wait another frame.
   else if (chk.done) {
      m.busy = undefined;
      return;
   }
   // Check width/height constraints.
   else {
      if (chk.cx !== undefined) {
         if (rc.width !== chk.rc.width) {
            m.x = chk.cx;
            m.minX = m.maxX = undefined;
            w.modified.size = true;
         }
         else {
            // Constrained by min/max width, reset to previous position/size.
            r.style.left = (chk.rc.left - m.px) + 'px';
            chk.p.style.width = chk.prc.width + 'px';
            if (m.minX === undefined && m.maxX === undefined) {
               if (chk.cx < m.x) {
                  m.minX = m.x;
                  m.maxX = undefined;
               }
               else {
                  m.maxX = m.x;
                  m.minX = undefined;
               }
            }
         }
      }
      if (chk.cy !== undefined) {
         if (rc.height !== chk.rc.height) {
            m.y = chk.cy;
            m.minY = m.maxY = undefined;
            w.modified.size = true;
         }
         else {
            // Constrained by min/max height, reset to previous position/size.
            r.style.top = (chk.rc.top - m.py) + 'px';
            chk.p.style.height = chk.prc.height + 'px';
            if (m.minY === undefined && m.maxY === undefined) {
               if (chk.cy < m.y) {
                  m.minY = m.y;
                  m.maxY = undefined;
               }
               else {
                  m.maxY = m.y;
                  m.minY = undefined;
               }
            }
         }
      }
      chk.done = true;
   }
   eq.Dom.onNextFrame(w, qi);
};
/** @preserve $Id: 10-popup.js,v 28.16 2021/09/23 13:24:33 mike Exp $
*//*======================================================================
   class PopupBox extends Window
   POPUP BOX class.
========================================================================*/

eq.c.Window.subClass('PopupBox', function(pb, parent) {
   if (eq.c.PopupBox.singleton)
      throw new Error("BUG: POPUP BOX singleton exists");
   var
      dm = eq.Dom,
      pp = document.createElement('DIV'),
      msg, btn, btnDefault = pb.btnDefault;
   pp.className = 'eq-popup-box';

   // Message.
   msg = pb.msg.split("\n");
   for (var i = 0, l = msg.length; i < l; i++) {
      el = document.createElement('P');
      dm.setContent(el, msg[i]);
      pp.appendChild(el);
   }

   // Buttons.
   btn = document.createElement('DIV');
   btn.className = 'eq-pushbutton';
   btnDefault = btnDefault > 0 ? btnDefault - 1 : 0;
   for (var i = 0, l = pb.btn.length; i < l; i++) {
      el = document.createElement('BUTTON');
      el.type = 'button';
      dm.setContent(el, pb.btn[i], true);
      if (i === btnDefault)
         el.classList.add('eq-focused');
      btn.appendChild(el);
   }
   pp.appendChild(btn);

   eq.c.Window.call(this, eq.c.Window.setupFrame(pp), parent);
   this.whenVisible = { cb : eq.c.PopupBox.setFocus };
   this.setTitle(pb.title ? pb.title : "Eloquence");
   this.setPos(-1, -1);
   this.activate();
   eq.c.PopupBox.singleton = this;
});

eq.c.PopupBox.setFocus = function(w) {
   w.root.querySelector('.eq-pushbutton>.eq-focused').focus();
};

/*------------------------------------------------------------------------
   static PopupBox.singleton
   Global POPUP BOX instance.
------------------------------------------------------------------------*/

eq.c.PopupBox.singleton = null;

/*------------------------------------------------------------------------
   PopupBox.close()
   Close POPUP BOX, submit event.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.close = function(e) {
   if (this !== eq.c.PopupBox.singleton)
      throw new Error("BUG: unexpected POPUP BOX invocation");
   if (e !== undefined) {
      var t = e.target;
      if (   eq.c.Window.fromTarget(t) === this
          && this.onMouseDownOrClick(e)) {
         e.stopPropagation();
         return false;
      }

      if (e.type !== 'click')
         return false;

      var i = this.buttonIndex(t);
      if (i === -1) {
         // Not a button.
         return false;
      }

      // Notify API thread.
      eq.app.api.postMessage({
         o : eq.ApiOp.onPopupEvent,
         d : i + 1
      });

      e.stopPropagation();
   }

   eq.c.Window.prototype.close.call(this);
   eq.c.PopupBox.singleton = null;
   return true;
};

/*------------------------------------------------------------------------
   PopupBox.onKey()
   POPUP BOX keyboard event handler.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.onKey = function(e) {
   if (this !== eq.c.PopupBox.singleton)
      throw new Error("BUG: unexpected POPUP BOX invocation");
   var t = e.target;
   if (   this.isEvent(e)
       || t.classList.contains('eq-idle')
       || t === document.body) {
      var i = 0;
      switch (e.key) {
         case 'Enter':
         case ' ':
         case 'Spacebar': // IE, Firefox <= 37
            if (this.buttonIndex(t) !== -1)
               eq.Dom.fireClick(t);
            break;
         case 'Tab':
            i = e.shiftKey ? -1 : 1;
            break;
         case 'ArrowUp':
         case 'Up':
         case 'ArrowLeft':
         case 'Left':
            i = -1;
            break;
         case 'ArrowDown':
         case 'Down':
         case 'ArrowRight':
         case 'Right':
            i = 1;
            break;
         default:
            return false;
      }

      if (i !== 0) {
         var
            btn = this.buttonIndex(t),
            el = this.root.querySelector(
               '.eq-popup-box>.eq-pushbutton').childNodes,
            l = el.length;
         if ((btn += i) < 0)
            btn = l - 1;
         else if (btn >= l)
            btn = 0;
         for (i = 0; i < l; i++)
            el[i].classList.remove('eq-focused');
         el[btn].classList.add('eq-focused');
         el[btn].focus();
      }

      eq.c.App.debounceKey(e.key);
      return true; // Consume event.
   }

   return false;
};

/*------------------------------------------------------------------------
   PopupBox.isEvent()
   Check if event applies to POPUP BOX.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.isEvent = function(e) {
   var el = e.target, r = this.root;
   while (el && el !== document && el !== r)
      el = el.parentNode;
   return el === r;
};

/*------------------------------------------------------------------------
   PopupBox.buttonIndex()
   Get POPUP BOX button index of specified target.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.buttonIndex = function(t) {
   if (t) {
      var pb;
      while ((pb = t.parentNode) && pb !== document && pb !== this.root) {
         var cl = pb.classList;
         if (cl && cl.contains('eq-pushbutton')) {
            var n = this.root.querySelector(
               '.eq-popup-box>.eq-pushbutton').childNodes;
            for (var i = 0, l = n.length; i < l; i++)
               if (t === n[i])
                  return i;
            break;
         }
         t = pb;
      }
   }
   return -1;
};

/** @preserve $Id: 99-main.js,v 28.58 2021/09/23 13:24:47 mike Exp $
*//*======================================================================
   class App extends Obj
   Main thread application class.
========================================================================*/

eq.c.Obj.subClass('App', function() {
   eq.c.Obj.call(this);

/* NOTYET

   // DLG anchor elements.
   // If not empty, dialog is anchored on first element.
   // If first element is null, dialog is anchored on body element.
   // Additional anchors are used to anchor dialog children by 'cssid'.
   this.dlgAnchor = [];
*/

   // Dialog map.
   this.dlg = new Map();

   // Current dialog.
   this.currDlg = null;

   // Focus.
   this.focus = null;
   this.pendingFocus = null;

   // Changed elements.
   this.changed = new Map();

   // Application is idle (not waiting for user input).
   this.idle = true;

   // Layout grid.
   this.grid = {
      // wr : Width reference 1ch
      // wp : Width pixels
      // wf : Width ch factor
      // hr : Height reference 1em
      // hp : Height pixels
      // hf : Height em factor
   };

   // Check page URL for query string app= parameter.
   var url = window.location.href, app;
   try {
      app = /[?&]app=([^&#]+)/i.exec(url)[1];
   } catch (e) {}
   if (app) {
      // Query string app= parameter specified, override application.
      window.sessionStorage.setItem('eq-webdlg.app', app);
      // Initialize user agent tag and session id.
      window.sessionStorage.removeItem('eq-webdlg.uat');
      window.sessionStorage.removeItem('eq-webdlg.sid');
      // Reload page wo/ query string app= parameter.
      window.location.replace(eq.Dom.updateQueryString(url, 'app'));
      // Try to minimize flicker before page is reloaded.
      document.body.style.display = 'none';
      this.redirected = true;
   }

   // Decode script query string.
   var
      uri, host, port,
      re = new RegExp('^(http[s]?://)([^/]+)(/[^?#]+/)[^?#]*[?#]?.*$', 'i');
   try {
      // Obtain script base URI from script URI.
      uri = re.exec(document.querySelector('script#eq-webdlg').src);
      this.scriptBase = uri[1] + uri[2] + uri[3];
   }
   catch (e) {
      throw new Error("WEBDLG: <script id=\"eq-webdlg\"> not found");
   }

   try {
      // Optional query string app= parameter.
      this.app = /[?&]app=([^&#]+)/i.exec(uri[0])[1];
   } catch (e) {}

   try {
      // Optional query string host= parameter.
      host = /[?&]host=([^&#]+)/i.exec(uri[0])[1];
   } catch (e) {}

   try {
      // Optional query string port= parameter.
      port = /[?&]port=([^&#]+)/i.exec(uri[0])[1];
   } catch (e) {}

   // Compose base and WebSocket URIs.
   uri = re.exec(location.href);
   host = host ? host : uri[2];
   if (port) {
      try {
         host = /^(.+:)[^:]*$/.exec(host)[1] + port;
      }
      catch (e) {
         host += ':' + port;
      }
   }
   this.appBase = uri[1] + host + '/';
   this.appWS = 'ws' + uri[1].substring(4).toLowerCase() + host + '/';
});

/*------------------------------------------------------------------------
   static App.doubleClickTimeout
   Max. time in millseconds after 'mouseup'
   before next 'mousedown' on same target.
------------------------------------------------------------------------*/

eq.c.App.doubleClickTimeout = 250;

/*------------------------------------------------------------------------
   static App.start()
   Start application.
------------------------------------------------------------------------*/

eq.c.App.start = function(e) {
   var self = eq.app, appOverride;

   // Deactivate 'eq-start' elements.
   var el = document.body.getElementsByClassName('eq-start');
   for (var i = 0, l = el.length; i < l; i++) {
      // Check for 'data-app' application override.
      if (e !== undefined && e.target === el[i] && ('app' in el[i].dataset))
         appOverride = el[i].dataset.app;

      // Remove 'click' handler.
      el[i].removeEventListener('click', eq.c.App.start);

      // Make disabled.
      el[i].classList.add('eq-disabled');
      if (el[i].disabled !== undefined)
         el[i].disabled = true;
   }

/* NOTYET

   // Find 'eq-anchor' elements if any.
   var
      el = document.body.getElementsByClassName('eq-anchor'),
      l = el.length;
   if (l) {
      // First, find 'eq-dialog' anchor if present.
      for (var i = 0, iDialog; i < l; i++)
         if (el[i].id === 'eq-dialog') {
            iDialog = i;
            break;
         }

      // Set first anchor.
      self.dlgAnchor.push(iDialog !== undefined ? 'eq-dialog' : null);

      // Next, find additional anchors if any.
      for (var i = 0; i < l; i++)
         if (i !== iDialog && el[i].id)
            self.dlgAnchor.push(el[i].id);
   }
*/

   // Add 'eq-webdlg' document class.
   document.documentElement.classList.add('eq-webdlg');

   // Notify API thread to start application.

   if (self.currApp === undefined)
   {
      if (appOverride !== undefined)
         self.currApp = appOverride;
      else {
         if (!self.app)
            throw new Error("Cannot start, application name undefined");
         self.currApp = self.app;
      }
   }

   // Persistently store current application
   // so it can be restarted if document is reloaded.
   window.sessionStorage.setItem('eq-webdlg.app', self.currApp);

   self.api.postMessage({
      o   : eq.ApiOp.openWebSocket,
      app : self.currApp,
      url : self.appWS,
      sid : window.sessionStorage.getItem('eq-webdlg.sid')
   });
};

/*------------------------------------------------------------------------
   static App.prepareStart()
   return: true if caller may immediately start application
------------------------------------------------------------------------*/

eq.c.App.prepareStart = function(restart) {
   var self = eq.app;

   // Restart application?
   if (restart === undefined) {
      // Invoked when DOM is ready.
      if (self.currApp = window.sessionStorage.getItem('eq-webdlg.app')) {
         // Yes, done.
         this.start();
         return false;
      }
   }
   else {
      // Application has stopped, clear session storage.
      self.updateSessionList(true);
      var sid, opn;
      if (   (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
          && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))
          && sid !== opn) {
         // Secondary session, close window.
         window.close();
      }
      window.sessionStorage.removeItem('eq-webdlg.app');
      window.sessionStorage.removeItem('eq-webdlg.uat');
      window.sessionStorage.removeItem('eq-webdlg.sid');
      window.sessionStorage.removeItem('eq-webdlg.opn');
   }
   if (restart) {
      // Invoked when application has stopped and may restart.
      if (   self.currApp !== undefined
          && document.documentElement.classList.contains('eq-restart')) {
         // Yes, done.
         this.start();
         return false;
      }
   }
   // No, invalidate current application.
   self.currApp = undefined;

   // Find 'eq-start' elements if any.
   var
      el = document.body.getElementsByClassName('eq-start'),
      l = el.length;
   if (!l) {
      // No 'eq-start' elements,
      // caller may immediately start application.
      return true;
   }

   // Setup 'eq-start' click handlers.
   for (var i = 0; i < l; i++) {
      el[i].addEventListener('click', this.start);
      el[i].classList.remove('eq-disabled');
      if (el[i].disabled !== undefined)
         el[i].disabled = false;
   }

   return false;
};

/*------------------------------------------------------------------------
   static App.stopped()
   Application stopped.
------------------------------------------------------------------------*/

eq.c.App.stopped = function() {
   var self = eq.app;

   // Session still alive? Wait for subsequent 'pagehide' event.
   self.stopped = window.setTimeout(eq.c.App.onStopped, 500);

   // Remove 'eq-webdlg' document class.
   document.documentElement.classList.remove('eq-webdlg');

   // Close POPUP BOX.
   if (self.popupBox !== undefined) {
      self.popupBox.close();
      self.popupBox = undefined;
   }

   // Clear dialog map and current dialog.
   self.dlg.forEach(eq.c.App.cbCloseDlg);
   self.dlg.clear();
   self.currDlg = null;

   // Clear focus.
   self.focus = null
   self.pendingFocus = null;

   // Clear changed elements.
   self.changed.clear();

/* NOTYET

   // Clear 'eq-anchor' elements.
   self.dlgAnchor.length = 0;
*/

   // Clear pending 'mousedown' and 'click' events.
   self.pendingMouseDown = undefined;
   self.pendingClick = undefined;

   if (self.capture !== undefined) {
      // Clear pending capture.
      self.endCapture();
   }

   // Close overlay.
   self.closeOverlay();

   // Clear pending 'once' rule.
   self.onceRule = undefined;
   self.onceId = undefined;

   // Clear layout mode.
   self.layoutMode = undefined;

   // Clear pending 'notified' status.
   self.notified = undefined;

   // Application is idle.
   self.idle = true;
};

eq.c.App.onStopped = function() {
   var self = eq.app;
   self.stopped = undefined;
   // Application stopped, no subsequent 'pagehide' event.
   // Clear session storage, session probably closed by server.
   self.updateSessionList(true);
   window.sessionStorage.removeItem('eq-webdlg.app');
   window.sessionStorage.removeItem('eq-webdlg.uat');
   window.sessionStorage.removeItem('eq-webdlg.sid');
   window.sessionStorage.removeItem('eq-webdlg.opn');
};

eq.c.App.cbCloseDlg = function(d) {
   d.close();
};

/*------------------------------------------------------------------------
   static App.domReady()
   Continue main thread initialization when DOM is ready.
------------------------------------------------------------------------*/

eq.c.App.domReady = function(e) {
   var app = eq.c.App;

   // Save original page title.
   app.title = document.title;

   // IE: suppress default help action on F1 key.
   if (window.onhelp !== undefined) {
      window.addEventListener('help', eq.c.App.onHelp, true);
      app.onHelpInstalled = true;
   }
   if (document.onhelp !== undefined) {
      document.addEventListener('help', eq.c.App.onHelp, true);
      app.onHelpInstalled = true;
   }

   // Notify API thread when Document is complete.
   if (document.readyState === 'complete')
      app.documentComplete();
   else
      window.addEventListener('load', app.documentComplete);

   // Prepare starting application.
   if (app.prepareStart()) {
      // Immediately start application.
      app.start();
   }

   // This method/listener is now obsolete.
   if (e !== undefined)
      window.removeEventListener('DOMContentLoaded', app.domReady);
   delete app.domReady;
};

/*------------------------------------------------------------------------
   static App.documentComplete()
   Notify API thread when document is complete.
------------------------------------------------------------------------*/

eq.c.App.documentComplete = function(e) {
   var app = eq.c.App, self = eq.app;

   // Calculate grid.
   var et = document.createElement('P');
   et.style.position = 'absolute';
   et.style.visibility = 'hidden';
   et.style.zIndex = eq.Dom.zIndex(document.body) - 1;
   et.style.top = et.style.left = '0';

   // Width/height reference 1ch/1em.
   var el = et.cloneNode(false);
   el.style.width = '1ch';
   el.style.height = '1em';
   document.body.appendChild(el);
   var rc = el.getBoundingClientRect(), w, h;
   self.grid.wr = w = rc.width;
   self.grid.hr = h = rc.height;
   document.body.removeChild(el);

   // Letter 'X' width/height.
   el = et.cloneNode(false);
   el.textContent = "X";
   document.body.appendChild(el);
   rc = el.getBoundingClientRect();
   self.grid.wp = rc.width;
   self.grid.wf = rc.width / w;
   self.grid.hp = rc.height + h * 0.125;
   self.grid.hf = rc.height / h + 0.125;
   document.body.removeChild(el);

   // Notify API thread.
   self.api.postMessage({
      o : eq.ApiOp.documentComplete,
      d : self.grid
   });

   // This method/listener is now obsolete.
   if (e !== undefined)
      window.removeEventListener('load', app.documentComplete);
   delete app.documentComplete;
};

/*------------------------------------------------------------------------
   static App.apiStart()
   Start API thread.
------------------------------------------------------------------------*/

eq.c.App.apiStart = function() {
   var app = eq.c.App, self = eq.app;
   self.api = new Worker(self.scriptBase + "eq-webdlg-api.js");
   self.api.onmessage = app.apiReceived;
   self.api.onerror = app.apiError;
};

/*------------------------------------------------------------------------
   static App.apiReceived()
   Process message from API thread.
------------------------------------------------------------------------*/

eq.c.App.apiReceived = function(m) {
   var app = eq.c.App, self = eq.app, op = eq.MainOp, resume = false;
   for (var di = 0, dl = m.data.length; di < dl; di++) {
      var o = m.data[di].o, d = m.data[di].d;
      switch (o) {
         case op.ssnNewOrResume:
            if (!d.uat)
               throw new Error("apiReceived ssnNewOrResume:" +
                  " invalid, user agent tag undefined");
            window.sessionStorage.setItem('eq-webdlg.uat', d.uat);
            if (!d.sid)
               throw new Error("apiReceived ssnNewOrResume:" +
                  " invalid, session id undefined");
            window.sessionStorage.setItem('eq-webdlg.sid', d.sid);
            if (!window.sessionStorage.getItem('eq-webdlg.opn'))
               window.sessionStorage.setItem('eq-webdlg.opn', d.sid);
            window.name = 'eq-webdlg.' + d.sid;
            self.updateSessionList();
            continue;

         case op.applicationStopped: {
            // Application has stopped.
            var sid, opn;
            if (   (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
                && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))
                && sid !== opn) {
               // Secondary session, close window.
               window.close();
            }
            else {
               app.stopped();
               if (d) {
                  // Restart or prepare restarting.
                  app.prepareStart(true);
               }
            }
            continue;
         }

         case op.applicationFailed:
            // Prepare restarting if necessary.
            app.stopped();
            self.currApp = undefined;
            app.prepareStart(false);
            continue;

         case op.layout: {
            self.layoutMode = d;
            continue;
         }

         case op.cancelCurrentCall:
            if (self.currDlg !== null) {
               eq.Dom.closeMenu();
               self.setIdle(self.currDlg);
               self.currDlg.ddo = undefined;
               self.currDlg = null;
               self.focus = null;
               self.changed.clear();
            }
            else if (self.popupBox !== undefined) {
               self.popupBox.close();
               self.popupBox = undefined;
            }
            continue;

         case op.openUrl: {
            var sid = window.sessionStorage.getItem('eq-webdlg.sid');
            if (!sid)
               throw new Error("apiReceived openUrl: invalid, sid undefined");
            if (!d.name)
               throw new Error("apiReceived openUrl: invalid, name undefined");
            if (!window.open(d.url, 'eq-webdlg.' + sid + '.' + d.name))
               console.error("Failed to open window for URL '"
                  + d.url + "'");
            continue;
         }

         case op.startApp: {
            // Open new window, set query string app= parameter.
            var url = eq.Dom.updateQueryString(window.location.href, 'app', d);
            if (!window.open(url))
               console.error("Failed to open window for application '"
                  + d + "'");
            continue;
         }

         case op.activate: {
            app.activateSession(self.sessionByTitle(d));
            continue;
         }

         case op.tabOrderUpdated:
            if (self.currDlg !== null && m.data[di].i === self.currDlg.id)
               self.currDlg.tabOrder = d;
            continue;

         case op.notify:
            if (self.currDlg !== null && m.data[di].i === self.currDlg.id
                                      && self.currDlg.ddo === undefined) {
               self.notified = true;
               self.submit(document.body, { type : 'notify' });
            }
            continue;
      }

      if (self.currDlg === null || self.currDlg.ddo === undefined) {
         // Sanity check except when Dialog.do currently active.
         if (!self.idle)
            throw new Error("apiReceived op " + o + ": invalid, not idle");
         if (self.currDlg !== null)
            throw new Error("apiReceived op " + o +
               ": invalid, current dialog pending");
         if (self.focus !== null)
            throw new Error("apiReceived op " + o + ": invalid, focus pending");
         if (self.changed.size)
            throw new Error("apiReceived op " + o + ": invalid, " +
               self.changed.size + " changed elements pending");
      }
      if (self.popupBox !== undefined)
         throw new Error("apiReceived op " + o + ": invalid, POPUP BOX open");

      switch (o) {
         case op.resume:
            resume = true;
            break;

         case op.addControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (d.pa !== undefined) {
               // Add to specified parent.
               if (dlg === undefined)
                  throw new Error("apiReceived addControl: DLG " +
                     id + " does not exist");
               var
                  p = document.getElementById(d.pa),
                  el;
               if (!p)
                  throw new Error("apiReceived addControl: parent '" +
                     d.pa + "' does not exist");
               p = p.querySelector('.eq-container');
               if (!p)
                  throw new Error("apiReceived addControl: parent '" +
                     d.pa + "' container not found");
               el = dlg.addControl(d, p);
/* NOTYET
               // TODO: dlgAnchor
               if (self.dlgAnchor.indexOf(d.pa) !== -1) {
                  el.classList.add('eq-root');
                  dlg.root.push(el);
                  app.addRootListeners(el);
                  self.setIdle(dlg);
               }
*/
            }
            else {
               // Add to root.
               if (dlg !== undefined)
                  dlg.close();
               else
                  dlg = self.newDialog(id);
               dlg.addRoot(dlg.addControl(d));
               self.setIdle(dlg);
            }
            break;
         }

         case op.modifyControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived modifyControl: DLG " +
                  id + " does not exist");
            if (!d.id)
               throw new Error("apiReceived modifyControl: id not set");
            if (!(el = document.getElementById(d.id)))
               throw new Error("apiReceived modifyControl: id '" +
                  d.id + "' does not exist");
            dlg.modifyControl(d.id, el, d);
            break;
         }

         case op.deleteControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id),
               el = document.getElementById(d),
               i;
            if (dlg === undefined)
               throw new Error("apiReceived deleteControl: DLG " +
                  id + " does not exist");
            if (!el)
               throw new Error("apiReceived deleteControl: id '" +
                  d + "' does not exist");
            if (dlg.removeElement(el))
               self.closeDialog(dlg);
            break;
         }

         case op.invalidate: {
            for (var i = d.length; --i >= 0;) {
               var dlg = self.dlg.get(d[i]);
               if (dlg !== undefined)
                  self.closeDialog(dlg);
            }
            break;
         }

         case op.dlg_DRAW: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived DLG DRAW: DLG " +
                  id + " does not exist");
            // Make Dialog top and visible.
            self.setCurrDialog(dlg);
            if (dlg.ddo === undefined) {
               // Not Dialog.do, immediately reset currDlg.
               self.currDlg = null;
            }
            // Notify API thread on next event cycle.
            eq.Dom.onNextCycle(app.sendDlgDrawResponse);
            break;
         }

         case op.dlg_DO: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived DLG DO: DLG " +
                  id + " does not exist");
            // Make Dialog top and visible.
            self.setCurrDialog(dlg);
            if (d.f)
               self.setFocus(d.f);
            if (d.t !== undefined)
               dlg.tabOrder = d.t;
            // Activate dialog.
            self.setIdle();
            // Execute pending focus action if any.
            if (self.pendingFocus !== null) {
               var f = self.pendingFocus, act, el;
               self.pendingFocus = null;
               if (   f.id === self.focus
                   && (act = f.act)
                   && (el = document.getElementById(f.id)))
                  act(el);
            }
            if (dlg.ddo !== undefined) {
               // Dialog.do implicit DLG DRAW,
               // notify API thread on next event cycle.
               eq.Dom.onNextCycle(app.sendDlgDrawResponse);
            }
            else if (dlg.w !== null) {
               // Reset Window close button state if necessary,
               // act upon pending Window close request.
               var el = dlg.w.root.querySelector(
                  '.eq-window>.eq-pane>.eq-caption>.eq-close.eq-click');
               if (el)
                  el.classList.remove('eq-click');
               if (dlg.pendingClose)
                  eq.Dom.onNextCycle(app.submitWindowClose);
            }
            break;
         }

         case op.dlg_POPUP_BOX: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            self.popupBox = new eq.c.PopupBox(d,
               document.getElementById('eq-dialog'));
            break;
         }

         default:
            throw new Error("apiReceived invalid op: " + o);
      }
   }

   if (resume) {
      self.dialogsVisible();
      var dlg = self.topDialog();
      if (dlg && dlg.ddo !== undefined) {
         // Resuming Dialog.do, activate dialog.
         if (self.currDlg !== null)
            throw new Error("apiReceived resume: " +
               "invalid, current dialog pending");
         self.setCurrDialog(dlg);
         if (dlg.ddo.f)
            self.setFocus(dlg.ddo.f);
         if (dlg.ddo.t)
            dlg.tabOrder = dlg.ddo.t;
         self.setIdle();
      }
   }
};

eq.c.App.sendDlgDrawResponse = function() {
   eq.app.api.postMessage({ o : eq.ApiOp.onDraw });
};

eq.c.App.submitWindowClose = function() {
   var
      self = eq.app,
      d = self.currDlg;
   if (d !== null && d.pendingClose) {
      d.pendingClose = undefined;
      self.submit(document.body, { type : 'close' });
   }
};

/*------------------------------------------------------------------------
   static App.apiError()
   Handle API thread error.
------------------------------------------------------------------------*/

eq.c.App.apiError = function(e) {
/* NOTYET

   var app = eq.c.App, self = eq.app;
   // Terminate and restart API thread.
   self.api.terminate();
   app.apiStart();
   // Stop application, prepare restarting if necessary.
   app.stopped();
   self.currApp = undefined;
   app.prepareStart(false);
*/
   // Issue error.
   throw new Error("WEBDLG API thread failed in line " +
      e.lineno + ": " + e.message);
};

/*------------------------------------------------------------------------
   static App.onEvent()
   Root element event handler.
------------------------------------------------------------------------*/

eq.c.App.onEvent = function(e) {
   var
      self = eq.app,
      dm = eq.Dom,
      md = e.type === 'mousedown';
   if (md) {
      // 'mousedown', invalidate last key.
      self.lastKey = undefined;
   }

   if (self.idle) {
      if (e.target.classList.contains('eq-idle')) {
         // Event on idle pane, set wait cursor.
         e.target.style.cursor = 'wait';
         dm.consumeEvent(e);
      }
      return;
   }

   var
      el = dm.eqControl(e.target), // Obtain 'eq-control' element.
      mdFocus = false,
      moEvent = false;
   if (el) {
      if (el.classList.contains('eq-disabled')) {
         dm.consumeEvent(e);
         if (self.menuOpen)
            dm.onNextCycle(dm.closeMenu);
         return;
      }

      var id = el.id, dlg = self.currDlg, ed;
      if (!dlg)
         throw new Error("BUG: onEvent no current dialog");
      if ((ed = dlg.eds.get(id)) !== undefined) {

         // 'mousedown'?
         if (md) {
            mdFocus = el.classList.contains('eq-focus');
            self.pendingMouseDown = { id : id, t : e.target };
         }
         else if (e.type === 'focusin') {
            // About to receive focus.
            if (el.classList.contains('eq-focus')) {
               // Set focus if different from current focus.
               if (id !== self.focus)
                  self.setFocus(id, el);
            }
            else {
               // Target not focusable, ignore event.
               dm.consumeEvent(e);
            }
            return;
         }

         self.processEvent(e, id, el, ed);
      }

      moEvent = self.menuOpen && dm.isMenu(el);
   }

   if (md) {
      if (!mdFocus) {
         // 'mousedown', target not focusable.
         dm.consumeEvent(e);
      }
      if (self.menuOpen && !moEvent) {
         // Menu open, not a menu event, close menu on next event cycle.
         dm.onNextCycle(dm.closeMenu);
      }
   }
};

/*------------------------------------------------------------------------
   App.processEvent()
   Process change and action event.
------------------------------------------------------------------------*/

eq.c.App.prototype.processEvent = function(e, id, el, ed) {
   var type = e.type;

   // Change event?
   if (ed.c !== undefined && ed.c.indexOf(type) !== -1) {
      // Yes, enqueue change if not already in queue.
      var ch = this.changed.get(id), cb;
      if (ch === undefined)
         this.changed.set(id, [ type ]);
      else if (ch.indexOf(type) === -1)
         ch.push(type);
      // Invoke 'OnChange' handler if present.
      if ((cb = eq.OnChange.get(ed.cb)) !== undefined)
         if (!cb(id, el, e))
            return false;
   }

   // Action event?
   if (ed.a !== undefined && ed.a.indexOf(type) !== -1) {
      // Yes, invoke 'OnEvent' handler if present.
      var cb = eq.OnEvent.get(ed.cb), pn;
      if (cb !== undefined && !cb(id, el, e))
         return false;
      if (type === 'mousedown') {
         // 'mousedown', register pending 'click'.
         if ((pn = this.pendingClick) !== undefined) {
            if (ed.m === undefined)
               throw new Error("BUG: onEvent pending click id:" + pn.id +
                  " cnt:" + pn.cnt);
            this.stopPendingClickTimer();
            if (id !== pn.id) {
               // target different from already pending 'click' target,
               // submit previous event, ignore this event.
               eq.c.App.commitPendingClick();
            }
            return true;
         }
         var ea = { btn : e.buttons };
         if (ea.btn === undefined) {
            switch (e.button) {
               case 2: // Secondary button, usually right button.
                  ea.btn = 2;
                  break;
               case 1: // Auxiliary button, usually wheel/middle button.
                  ea.btn = 4;
                  break;
               default: // Assume main button, usually left button.
                  ea.btn = 1;
            }
         }
         this.pendingClick = {
            id : id,
            type : 'click',
            cnt : 0,
            ea : ea
         };
      }
      else if ((pn = this.pendingClick) !== undefined) {
         // 'click' pending, append event.
         if (pn.ev !== undefined)
            throw new Error("BUG: onEvent pending click id:" + pn.id +
               " and id:" + pn.ev.id + " type:" + pn.ev.type);
         pn.ev = {
            id : id,
            type : type,
            ea : e.eq
         };
      }
      else {
         // Submit event.
         this.submit(el, {
            id : id,
            type : type,
            ea : e.eq
         });
      }

      return true;
   }

   return false;
};

/*------------------------------------------------------------------------
   App.setChanged()
   Set element changed.
------------------------------------------------------------------------*/

eq.c.App.prototype.setChanged = function(id, ty) {
   var dlg = this.currDlg, ed;
   if (!dlg)
      throw new Error("BUG: onEvent no current dialog");
   if ((ed = dlg.eds.get(id)) !== undefined) {
      if (ed.c !== undefined && ed.c.indexOf(ty) !== -1) {
         var ch = this.changed.get(id);
         if (ch === undefined)
            this.changed.set(id, [ ty ]);
         else if (ch.indexOf(ty) === -1)
            ch.push(ty);
      }
   }
};

/*------------------------------------------------------------------------
   App.elementsChanged()
   Collect and return element changes.
------------------------------------------------------------------------*/

eq.c.App.prototype.elementsChanged = function(ea) {
   var
      dlg = eq.app.currDlg,
      w, r, p, m,
      changed = [{
         ty : eq.RsValType.idFocus,
         sv : this.focus !== null ? this.focus : ""
      }];
   if (!dlg)
      throw new Error("BUG: elementsChanged no current dialog");
   if (ea !== undefined) {
      // Dialog-specific event attributes.
      if (ea.key !== undefined) {
         changed.push({
            ty : eq.RsValType.ruleKey,
            iv : ea.key
         });
      }
      else {
         if (ea.btn !== undefined) {
            changed.push({
               ty : eq.RsValType.mouseBtns,
               iv : ea.btn
            });
         }
         if (ea.cnt !== undefined) {
            changed.push({
               ty : eq.RsValType.mouseCnt,
               iv : ea.cnt
            });
         }
         if (ea.mod !== undefined) {
            changed.push({
               ty : eq.RsValType.mouseMods,
               iv : ea.mod
            });
         }
      }
   }
   if (this.notified) {
      changed.push({
         ty : eq.RsValType.notified
      });
      this.notified = undefined;
   }
   if ((w = dlg.w) !== null && (p = (r = w.root).querySelector(
         '.eq-window>.eq-pane>.eq-dialog>.eq-pane'))) {
      if ((m = w.modified).maximized) {
         m.maximized = undefined;
         changed.push({
            ty : eq.RsValType.dialogMax,
            iv : r.classList.contains('eq-max') ? 1 : 0
         });
      }
      if (m.pos) {
         m.pos = undefined;
         changed.push({
            ty : eq.RsValType.dialogPos,
            x : r.offsetLeft,
            y : r.offsetTop
         });
      }
      if (m.size) {
         m.size = undefined;
         changed.push({
            ty : eq.RsValType.dialogSize,
            w : p.offsetWidth,
            h : p.offsetHeight
         });
      }
   }
   if (this.onceRule !== undefined) {
      changed.push({
         id : this.onceId,
         ty : eq.RsValType.ruleOnce,
         iv : this.onceRule
      });
      this.onceRule = undefined;
      this.onceId = undefined;
   }
   this.changed.forEach(eq.c.App.cbElementsChanged, { d: dlg, c : changed });
   this.changed.clear();
   return changed;
};

eq.c.App.cbElementsChanged = function(ty, id) {
   var
      dlg = this.d,
      ed, el, cb, v;
   if ((ed = dlg.eds.get(id)) === undefined)
      throw new Error("BUG: elementsChanged id:" + id +
         " not registered, type:" + ty);
   if ((cb = eq.Changed.get(ed.cb)) === undefined)
      throw new Error("BUG: elementsChanged id:" + id +
         " 'Changed' handler undefined, type:" + ty);
   el = document.getElementById(id);
   if (!el.classList.contains('eq-event'))
      el = el.querySelector('.eq-event');
   if (!el)
      throw new Error("BUG: cbElementsChanged id:" + id +
         " event element not found");
   if ((v = cb(id, el, ty)) !== null)
      this.c.push(v);
};

/*------------------------------------------------------------------------
   App.submit()
   Submit event.
------------------------------------------------------------------------*/

eq.c.App.prototype.submit = function(el, ev) {
   var
      dm = eq.Dom,
      app = eq.c.App,
      dlg = this.currDlg,
      ddo;
   dm.closeMenu();
   this.closeOverlay();
   if (this.capture !== undefined) {
      // Clear pending capture.
      this.endCapture();
   }
   if (el.classList.contains('eq-help')) {
      app.debounceLastKey();
      this.api.postMessage({
         o : eq.ApiOp.onHelp,
         d : el.id
      });
      return;
   }
   if ((ddo = dlg.ddo) !== undefined) {
      // Dialog.do event.
      dlg.ddo = undefined;
      var
         f = this.focus !== null ? this.focus : "",
         r;
      if (ddo.hide || ddo.show) {
         // Hide and/or show as specified.
         if (ddo.hide && (el = document.getElementById(ddo.hide)))
            el.style.display = 'none';
         if (ddo.show && (el = document.getElementById(ddo.show)))
            el.style.display = '';
      }
      else {
         // Hide Dialog.
         for (var i = dlg.root.length; --i >= 0;) {
            if ((r = dlg.root[i]).classList.contains('eq-dialog'))
               dlg.w.root.style.display = 'none';
            else
               r.style.display = 'none';
         }
      }
      this.currDlg = null;
      this.focus = null;
      this.changed.clear();
      this.setIdle(dlg);
      app.debounceLastKey();
      // Notify API thread.
      this.api.postMessage({
         o : eq.ApiOp.onDo,
         i : dlg.id,
         f : f
      });
      return;
   }
   // Special case for CheckBox, RadioButton,
   // enqueue change if not already in queue.
   if (   (   el.classList.contains('eq-checkbox')
           || el.classList.contains('eq-radiobutton'))
       && !this.changed.has(el.id))
      this.changed.set(el.id, 'input');
   // Notify API thread on next event cycle.
   this.setIdle(dlg);
   app.debounceLastKey();
   dm.onNextCycle(app.apiSubmit, ev);
};

eq.c.App.apiSubmit = function(ev) {
   var self = eq.app;
   ev.changed = self.elementsChanged(ev.ea);
   self.api.postMessage({
      o : eq.ApiOp.onEvent,
      d : ev
   });
   // Clear current dialog and focus.
   self.currDlg = null;
   self.focus = null;
};

/*------------------------------------------------------------------------
   static App.onGlobalKeyDown() .onGlobalKeyUp()
   Global keyboard event handler.
------------------------------------------------------------------------*/

eq.c.App.onGlobalKeyDown = function(e) {
   var self = eq.app, dm = eq.Dom;
   if (e.key === self.debounceKey) {
      eq.c.App.debounceKey(e.key);
      dm.consumeEvent(e);
      return;
   }

   if (self.popupBox !== undefined) {
      // POPUP BOX is open.
      if (self.popupBox.onKey(e))
         dm.consumeEvent(e);
      return;
   }

   // Current Dialog?
   var dlg = self.currDlg;
   if (dlg) {
      // Obtain target element, overlay or 'eq-control' or focus.
      var el, id, ed, cl;
      if (self.overlay !== undefined)
         el = self.overlay.el;
      else if (self.focus !== null)
         if (e.target === document.body || dm.eqControl(e.target))
            el = document.getElementById(self.focus);
      if (el) {
         if ((cl = el.classList).contains('eq-disabled')) {
            dm.consumeEvent(e);
            return;
         }
         if ((ed = dlg.eds.get(id = el.id)) === undefined)
            return;
         // Target is DLG element.
         if (self.idle) {
            // Ignore event while idle.
            // TODO: type-ahead
            dm.consumeEvent(e);
            return;
         }
      }

      // Navigation or function key?
      var fkey;
      switch (e.key) {
         case 'Enter':
            if (cl && cl.contains('eq-enter')) {
               // Enter key has special function.
               break;
            }
            if (dlg.fkey && dlg.fkey.length > 0 && dlg.fkey[0]) {
               // Dialog.cr function key.
               fkey = 0;
               break;
            }
            if (!cl || cl.contains('eq-noenter')) {
               // No focused element, or .tabonenter disabled.
               break;
            }
            // FALLTHROUGH
         case 'Tab':
            if (el) {
               self.lastKey = e.key;
               dm.consumeEvent(e);

               // Navigate to next/previous element in tab order.
               var tb = dlg.tabOrder;
               if (tb !== null) {
                  var i = tb.indexOf(id);
                  if (i !== -1) {
                     var
                        shift = e.shiftKey,
                        j = shift ? i - 1 : i + 1,
                        l = tb.length;
                     while (true) {
                        if (shift) {
                           if (j === -1)
                              j = l - 1;
                        }
                        else {
                           if (j === l)
                              j = 0;
                        }
                        if (j === i) {
                           // No other element found in tab order,
                           // submit 'blur' event if registered.
                           el = document.getElementById(id = tb[i]);
                           var cb = eq.FocusElement.get(ed.cb);
                           self.processEvent({
                              type : 'blur',
                              target : cb !== undefined ? cb(el) : el
                           }, id, el, ed);
                           break;
                        }

                        el = document.getElementById(id = tb[j]);
                        if (el && !el.classList.contains('eq-notab')) {
                           // Found next/previous element in tab order.
                           self.setFocus(id, el);
                           break;
                        }

                        if (shift)
                           j--;
                        else
                           j++;
                     }
                  }
               }
            }
            return;

         case 'F1':
            fkey = 1;
            if (eq.c.App.onHelpInstalled) {
               // IE: suppress default help action on F1 key.
               self.onHelpIgnore = true;
            }
            break;
         case 'F2':
            fkey = 2;
            break;
         case 'F3':
            fkey = 3;
            break;
         case 'F4':
            fkey = 4;
            break;
         case 'F5':
            fkey = 5;
            break;
         case 'F6':
            fkey = 6;
            break;
         case 'F7':
            fkey = 7;
            break;
         case 'F8':
            fkey = 8;
            break;
         case 'F9':
            fkey = 9;
            break;
         case 'F10':
            fkey = 10;
            break;
         case 'F11':
            fkey = 11;
            break;
         case 'F12':
            fkey = 12;
            break;
      }

      if (el && fkey === undefined) {
         var cb = eq.OnKey.get(ed.cb);
         if (cb !== undefined && cb(id, el, ed, e)) {
            dm.consumeEvent(e);
            return;
         }
      }

      if (fkey !== undefined) {
         // Function key?
         if (dlg.fkey && fkey < dlg.fkey.length) {
            var r = dlg.fkey[fkey];
            if (r) {
               self.lastKey = e.key;
               dm.consumeEvent(e);

               if (r > 0) {
                  // Submit function key rule.
                  self.onceRule = r;
                  self.submit(document.body, { type : 'key' });
               }
               else if (r === -1)
                  self.api.postMessage({
                     o : eq.ApiOp.onHelp,
                     d : el ? el.id : null
                  });

               return;
            }
         }

         self.onHelpIgnore = undefined;
      }
   }
};

eq.c.App.onGlobalKeyUp = function(e) {
   var self = eq.app;
   if (e.key === self.debounceKey) {
      window.clearTimeout(self.debounceKeyTimer);
      self.debounceKeyTimer = undefined;
      self.debounceKey = undefined;
      eq.Dom.consumeEvent(e);
   }
   if (e.key === self.lastKey)
      self.lastKey = undefined;
};

/*------------------------------------------------------------------------
   static App.debounceKey() .debounceLastKey()
   Debounce key utilities.
------------------------------------------------------------------------*/

eq.c.App.debounceKey = function(key) {
   var self = eq.app;
   self.debounceKey = key;
   if (self.debounceKeyTimer !== undefined)
      window.clearTimeout(self.debounceKeyTimer);
   self.debounceKeyTimer = window.setTimeout(this.onDebounceKeyTimer, 250);
   self.lastKey = undefined;
};

eq.c.App.debounceLastKey = function() {
   var key = eq.app.lastKey;
   if (key !== undefined)
      this.debounceKey(key);
};

eq.c.App.onDebounceKeyTimer = function() {
   var self = eq.app;
   // Timer has elapsed.
   self.debounceKeyTimer = undefined;
   self.debounceKey = undefined;
};

/*------------------------------------------------------------------------
   static App.help()
   IE: Selectively filter F1 function key default action.
------------------------------------------------------------------------*/

eq.c.App.onHelp = function(e) {
   var self = eq.app;
   if (self.onHelpIgnore) {
      // IE: suppress default help action on F1 key.
      self.onHelpIgnore = undefined;
      eq.Dom.consumeEvent(e);
      return false;
   }
   return true;
};

/*------------------------------------------------------------------------
   static App.onGlobalMouseDownOrClick()
   Global 'mousedown' or 'click' event handler.
------------------------------------------------------------------------*/

eq.c.App.onGlobalMouseDownOrClick = function(e) {
   var
      self = eq.app,
      dm = eq.Dom,
      md = e.type === 'mousedown';

   if (self.capture !== undefined) {
      // End capture.
      e.stopPropagation();
      self.endCapture();
   }

   // 'mousedown' on overlay?
   if (md && self.overlay !== undefined) {
      if (dm.eqOverlay(e.target)) {
         var
            el = self.overlay.el,
            id = el.id,
            dlg = self.currDlg,
            ed;
         if (!dlg)
            throw new Error(
               "BUG: onGlobalMouseDownOrClick overlay no current dialog");
         if ((ed = dlg.eds.get(id)) !== undefined)
            self.processEvent(e, id, el, ed);
      }
      else if (dm.eqControl(e.target) !== self.overlay.el) {
         // Clicked outside overlay, close.
         self.closeOverlay();
      }
   }

   // Close menu if clicked outside Dialog or MenuBar.
   if (self.menuOpen) {
      var
         el = e.target,
         cl, found;
      while (el && el !== document && (cl = el.classList)) {
         if (cl.contains('eq-dialog') || cl.contains('eq-menubar')) {
            found = true;
            break;
         }
         el = el.parentNode;
      }
      if (!found)
         dm.onNextCycle(dm.closeMenu);
   }

   // POPUP BOX.
   if (self.idle && self.popupBox !== undefined) {
      if (self.popupBox.isEvent(e)) {
         if (self.popupBox.close(e)) {
            self.popupBox = undefined;
            return;
         }
      }
      else if (!e.target.classList.contains('eq-idle'))
         return;

      if (md) {
         // 'mousedown', target not focusable.
         e.preventDefault();
      }
   }
   else {
      var w = eq.c.Window.fromTarget(e.target);
      if (w) {
         w.onMouseDownOrClick(e);
         if (md) {
            // 'mousedown', target not focusable.
            dm.consumeEvent(e);
         }
      }
   }
};

/*------------------------------------------------------------------------
   static App.onGlobalMouseUp()
   Global 'mouseup' event handler.
------------------------------------------------------------------------*/

eq.c.App.onGlobalMouseUp = function(e) {
   var self = eq.app;
   if (self.capture !== undefined) {
      // End capture.
      e.stopPropagation();
      self.endCapture();
   }

   if (self.idle) {
      // 'mouseup' while idle.
      self.pendingMouseDown = undefined;
      // 'click' must not be pending.
      if (self.pendingClick !== undefined) {
         var id = self.pendingClick.id;
         self.pendingClick = undefined;
         throw new Error("BUG: idle onGlobalMouseUp pending click id:" + id);
      }
      return;
   }

   var dm = eq.Dom, pn;
   if ((pn = self.pendingMouseDown) !== undefined) {
      // 'mouseup' while 'mousedown' pending.
      var t = pn.t;
      if (t === e.target) {
         // 'mouseup' target same as pending 'mousedown' target,
         // obtain 'eq-control' element.
         var
            id = pn.id,
            dlg = self.currDlg,
            el = dm.eqControl(t);
         if (!dlg)
            throw new Error("BUG: onGlobalMouseUp no current dialog");
         // 'click' committed.
         self.processEvent({
            type : 'click',
            target : t
         }, id, el, dlg.eds.get(id));
         if (el.classList.contains('eq-focus')) {
            // Element focusable, set focus.
            self.setFocus(id, el);
         }
      }
      else if (self.menuOpen && dm.isMenu(t = e.target)) {
         // menu is open, 'mouseup' target is menu element,
         // obtain 'eq-control' element.
         var
            dlg = self.currDlg,
            el = dm.eqControl(t),
            id = el.id;
         if (!dlg)
            throw new Error("BUG: onGlobalMouseUp no current dialog");
         // 'click' committed.
         self.processEvent({
            type : 'click',
            target : t
         }, id, el, dlg.eds.get(id));
      }

      self.pendingMouseDown = undefined;

      // Close menu if item click not submitted.
      if (self.menuOpen && dm.isMenu(t = e.target)) {
         var el = dm.eqControl(t);
         if (el && el.classList.contains('eq-item'))
            dm.onNextCycle(dm.closeMenu);
      }
   }

   if ((pn = self.pendingClick) !== undefined) {
      // 'mouseup' while 'click' pending.
      self.stopPendingClickTimer();
      // Obtain 'eq-control' element.
      var
         dlg = self.currDlg,
         el = dm.eqControl(e.target),
         id, ed;
      if (!dlg)
         throw new Error("BUG: onGlobalMouseUp pendingClick no current dialog");
      if (el && (id = el.id) === pn.id) {
         // 'mouseup' target same as pending 'click' target,
         // obtain event attributes btn, cnt, mod.
         if (e.eq !== undefined)
            pn.ea = e.eq;
         else {
            var
               ea = pn.ea = pn.ea || {},
               mod = 0;
            ea.cnt = pn.cnt ? pn.cnt + 1 : 1;
            if (e.shiftKey)
               mod += 1;
            if (e.ctrlKey)
               mod += 2;
            if (e.metaKey)
               mod += 4;
            if (e.altKey)
               mod += 8;
            if (mod)
               ea.mod = mod;
         }
         if ((ed = dlg.eds.get(id)) !== undefined && ed.m !== undefined)
         {
            // Check number of mouse clicks.
            if (pn.cnt === 0) {
               // First click, wait for second click.
               pn.cnt = 1;
               var app = eq.c.App;
               self.commitPendingClickTimer =
                  window.setTimeout(app.commitPendingClick,
                                    app.doubleClickTimeout);
               return;
            }
         }
         // 'click' committed, submit event.
         self.submit(el, pn);
      }
      else if ((pn = pn.ev) !== undefined) {
         // 'mouseup' target different from pending 'click' target,
         // submit pending 2nd event.
         if (el = document.getElementById(pn.id))
            self.submit(el, pn);
      }

      self.pendingClick = undefined;
   }
};

eq.c.App.commitPendingClick = function() {
   var self = eq.app, pn;
   self.commitPendingClickTimer = undefined;
   if ((pn = self.pendingClick) !== undefined) {
      var id = pn.id, dlg, ed, el;
      if ((dlg = self.currDlg) && (ed = dlg.eds.get(id)) !== undefined) {
         if (ed.m === undefined)
            throw new Error(
               "BUG: commitPendingClick undefined pending click id:" + id +
               " cnt:" + pn.cnt);
         if (pn.cnt >= ed.m) {
            // 'click' committed, submit event.
            if (el = document.getElementById(id))
               self.submit(el, pn);
         }
         else if ((pn = pn.ev) !== undefined) {
            // submit pending 2nd event.
            if (el = document.getElementById(pn.id))
               self.submit(el, pn);
         }
      }
      self.pendingClick = undefined;
   }
};

eq.c.App.prototype.stopPendingClickTimer = function() {
   if (this.commitPendingClickTimer !== undefined) {
      window.clearTimeout(this.commitPendingClickTimer);
      this.commitPendingClickTimer = undefined;
   }
};

/*------------------------------------------------------------------------
   App.moveCapture()
   Begin 'mousemove' capture.
------------------------------------------------------------------------*/

eq.c.App.prototype.moveCapture = function(arg, cb, done) {
   if (this.capture !== undefined)
      this.endCapture();
   this.capture = { arg : arg, cb : cb, done : done };
   window.addEventListener('mousemove', eq.c.App.onCapture);
};

eq.c.App.prototype.endCapture = function() {
   var c = this.capture;
   window.removeEventListener('mousemove', eq.c.App.onCapture);
   this.capture = undefined;
   if (c.done)
      c.done(c.arg);
};

eq.c.App.onCapture = function(e) {
   e.stopPropagation();
   var c = eq.app.capture;
   if (!c)
      window.removeEventListener('mousemove', eq.c.App.onCapture);
   else if (c.cb)
      c.cb(e, c.arg);
};

/*------------------------------------------------------------------------
   static App.onGlobalResize()
   Global 'resize' event handler.
------------------------------------------------------------------------*/

eq.c.App.onGlobalResize = function(e) {
   if (e.target === window) {
      var self = eq.app;
      self.closeOverlay();
      if (self.capture !== undefined) {
         // Clear pending capture.
         self.endCapture();
      }
   }
};

/*------------------------------------------------------------------------
   static App.onGlobalBlur()
   Global 'blur' event handler.
------------------------------------------------------------------------*/

eq.c.App.onGlobalBlur = function(e) {
   if (e.target === window) {
      var self = eq.app;
      eq.Dom.closeMenu();
      self.closeOverlay();
      if (self.capture !== undefined) {
         // Clear pending capture.
         self.endCapture();
      }
   }
};

/*------------------------------------------------------------------------
   static App.onGlobalPageHide()
   Global 'pagehide' event handler, fired when leaving current page.
------------------------------------------------------------------------*/

eq.c.App.onGlobalPageHide = function(e) {
   var self = eq.app;
   if (self.stopped !== undefined) {
      // Application stopped because user closed window or moved away.
      // Restart session later if possible, see eq.c.App.onStopped().
      window.clearTimeout(self.stopped);
      self.stopped = undefined;
   }
   self.updateSessionList(true);
};

/*------------------------------------------------------------------------
   static App.onStorageModified()
   Global 'storage' event handler, storage modified in other document.
   Updates session list from localStorage if necessary.
------------------------------------------------------------------------*/

eq.c.App.onStorageModified = function(e) {
   if (e.storageArea === window.localStorage) {
      switch (e.key) {
         case 'eq-webdlg.sls': {
            // Session list.
            var self = eq.app;
            if (!e.newValue) {
               // Session list deleted.
               self.sls = undefined;
            }
            else {
               // Session list modified.
               try {
                  var
                     nsls = JSON.parse(e.newValue),
                     csls = self.sls;
                  if (!csls || nsls.ts >= csls.ts) {
                     // No current session list,
                     // or modified list is newer, take modified list.
                     self.sls = nsls;
                  }
               }
               catch (e) {
                  console.log("Failed to parse session list, " + e.message);
                  self.sls = undefined;
                  window.localStorage.removeItem('eq-webdlg.sls');
               }
            }
            self.updateSessionList();
            break;
         }

         case 'eq-webdlg.act': {
            // Activate session.
            var sid;
            if (   e.newValue
                && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
                && e.newValue === sid) {
               window.focus();
               window.localStorage.removeItem('eq-webdlg.act');
            }
            break;
         }
      }
   }
};

/*------------------------------------------------------------------------
   static App.activateSession()
   Activate other session.
------------------------------------------------------------------------*/

eq.c.App.activateSession = function(s) {
   var sid, w;
   if (   s
       && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
       && s !== sid
       && (w = window.open('', 'eq-webdlg.' + s))) {
      // Try to activate session window.
      w.focus();
      window.localStorage.setItem('eq-webdlg.act', s);
   }
};

/*------------------------------------------------------------------------
   App.sessionTitles()
   Obtain titles of all other sessions.
------------------------------------------------------------------------*/

eq.c.App.prototype.sessionTitles = function() {
   var t = [], sls = this.sls, uat, opn, sn, s;
   if (   sls !== undefined
       && (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
       && sls.ua == uat
       && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))) {
      for (var i = 0, l = (sn = sls.sn).length; i < l; i++)
         if (opn === (s = sn[i]).o && s.t)
            t.push(s.t);
   }
   return t;
};

/*------------------------------------------------------------------------
   App.sessionByIndex()
   Obtain other session by index.
------------------------------------------------------------------------*/

eq.c.App.prototype.sessionByIndex = function(si) {
   var sls = this.sls, uat, opn, sn, s;
   if (   sls !== undefined
       && (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
       && sls.ua == uat
       && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))) {
      for (var i = 0, l = (sn = sls.sn).length; i < l; i++)
         if (opn === (s = sn[i]).o && s.t && si-- === 0)
            return s.s;
   }
   return null;
};

/*------------------------------------------------------------------------
   App.sessionByTitle()
   Obtain other session by title.
------------------------------------------------------------------------*/

eq.c.App.prototype.sessionByTitle = function(t) {
   var sls = this.sls, uat, sid, opn, sn, s;
   if (   sls !== undefined
       && (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
       && sls.ua == uat
       && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
       && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))) {
      for (var i = 0, l = (sn = sls.sn).length; i < l; i++)
         if (opn === (s = sn[i]).o && sid !== s.s && t === s.t)
            return s.s;
   }
   return null;
};

/*------------------------------------------------------------------------
   App.updateSessionList()
   Update session list.
------------------------------------------------------------------------*/

eq.c.App.prototype.updateSessionList = function(closing) {
   var uat, sid, opn, sls, sn;
   if (   !closing
       && (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
       && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
       && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))) {
      if ((sls = this.sls) === undefined) {
         try {
            if (sls = JSON.parse(window.localStorage.getItem('eq-webdlg.sls')))
               this.sls = sls;
         } catch (e) {}
      }
      if (!sls || sls.ua !== uat) {
         // No session list or different user agent tag.
         this.sls = sls = { ua: uat, sn : [] };
      }
      var i, l = (sn = sls.sn).length, t = document.title, s, modified;
      for (i = 0; i < l; i++) {
         if (sid === (s = sn[i]).s) {
            if (s.o !== opn || s.t !== t) {
               s.o = opn;
               s.t = t;
               modified = true;
            }
            break;
         }
      }
      if (i === l) {
         sn.push({
            s : sid,
            o : opn,
            t : t
         });
         modified = true;
      }
      if (modified) {
         sls.ts = Date.now();
         window.localStorage.setItem('eq-webdlg.sls', JSON.stringify(sls));
      }
   }
   else {
      this.sls = undefined;
      if (   (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
          && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))) {
         try {
            sls = JSON.parse(window.localStorage.getItem('eq-webdlg.sls'));
            if (sls.ua === uat)
               for (var i = 0, l = (sn = sls.sn).length; i < l; i++)
                  if (sid === sn[i].s) {
                     sn.splice(i, 1);
                     sls.ts = Date.now();
                     window.localStorage.setItem(
                        'eq-webdlg.sls', JSON.stringify(sls));
                     break;
                  }
         } catch (e) {}
      }
   }
};

/*------------------------------------------------------------------------
   static App.addRootListeners() .removeRootListeners()
   Add/remove root element event listeners.
------------------------------------------------------------------------*/

eq.c.App.rootEventTypes = [
   'change',
   'focusin',
   'input',
   'mousedown'
];

eq.c.App.addRootListeners = function(el) {
   var ln = this.onEvent, types = this.rootEventTypes;
   for (var i = 0, l = types.length; i < l; i++)
      el.addEventListener(types[i], ln, true);
};

eq.c.App.removeRootListeners = function(el) {
   var ln = this.onEvent, types = this.rootEventTypes;
   for (var i = 0, l = types.length; i < l; i++)
      el.removeEventListener(types[i], ln, true);
};

/*------------------------------------------------------------------------
   App.setFocus()
   Set focus by element id.
------------------------------------------------------------------------*/

eq.c.App.prototype.setFocus = function(id, el) {
   // Focus changed?
   if (id !== this.focus) {
      var
         dlg = this.currDlg,
         n, f, ed, cb;
      if (!dlg)
         throw new Error("BUG: setFocus no current dialog");

      // Remove current focus indicator.
      n = document.body.getElementsByClassName('eq-focused');
      for (var i = 0, l = n.length; i < l; i++)
         n[i].classList.remove('eq-focused');

      if ((f = this.focus) !== null) {
         // Focus change, check 'blur' event on current focus element.
         if ((ed = dlg.eds.get(f)) === undefined)
            throw new Error("setFocus: from '" + f + "' not registered");
         var e, from = document.getElementById(f);
         if (!from)
            throw new Error("setFocus: from '" + f + "' does not exist");
         if ((cb = eq.ResetFocus.get(ed.cb)) !== undefined)
            cb(from);
         cb = eq.FocusElement.get(ed.cb);
         if (this.processEvent({
            type : 'blur',
            target : cb !== undefined ? cb(from) : from
         }, f, from, ed)) {
            // Action event submitted, done.
            this.focus = id;
            return true;
         }
      }

      // Set new focus.
      if ((ed = dlg.eds.get(id)) === undefined)
         throw new Error("setFocus: to '" + id + "' not registered");
      if (el === undefined && !(el = document.getElementById(id)))
         throw new Error("setFocus: to '" + id + "' does not exist");
      el.classList.add('eq-focused');
      if ((cb = eq.FocusElement.get(ed.cb)) !== undefined)
         el = cb(el);
      if ((cb = eq.SetFocus.get(ed.cb)) !== undefined)
         cb(el);
      this.focus = id;
      // Set element focus on next frame.
      eq.Dom.onNextFrame(el, { cb : eq.c.App.setElementFocus});
   }
   return false;
};

eq.c.App.setElementFocus = function(el) {
   el.focus();
   // Verify element focus on next event frame.
   eq.Dom.onNextFrame(el, { cb : eq.c.App.verifyElementFocus});
};

eq.c.App.verifyElementFocus = function() {
   var f = eq.app.focus, el;
   if (f && (!(el = eq.Dom.eqControl(document.activeElement)) || el.id !== f))
      // Focus could not be set, set focus to body element.
      if ((el = document.activeElement) && el !== document.body)
         el.blur();
};

/*------------------------------------------------------------------------
   App.setIdle()
   Set application idle or active.
------------------------------------------------------------------------*/

eq.c.App.prototype.setIdle = function(dlg) {
   var app = eq.c.App;
   this.stopIdleTimer();
   if (dlg !== undefined) {
      // Idle.
      this.idle = true;
      // Add idle pane(s) if necessary.
      var added = false;
      for (var i = 0, l = dlg.root.length; i < l; i++) {
         var el = dlg.root[i];
         if (!el.querySelector('.eq-root>.eq-idle')) {
            app.createIdlePane(el);
            added = true;
         }
         // After 5 seconds, visualize idle state on newly added panes.
         if (added)
            this.idleTimer = window.setTimeout(app.onIdleTimer, 5000);
      }
   }
   // Delay setting active state until next event cycle.
   else
      window.setTimeout(app.setActive);
};

// Set active state.
eq.c.App.setActive = function() {
   var
      self = eq.app,
      dlg = self.currDlg,
      el;
   if (!dlg)
      throw new Error("BUG: setActive no current dialog");
   self.idle = false;
   // Remove idle pane(s) if any.
   for (var i = 0, l = dlg.root.length; i < l; i++)
      if (el = dlg.root[i].querySelector('.eq-root>.eq-idle'))
         el.parentNode.removeChild(el);
};

// Visualze idle state.
eq.c.App.onIdleTimer = function() {
   // Timer has elapsed.
   eq.app.idleTimer = undefined;
   // Add 'eq-fog' class to idle pane(s) if necessary.
   var el = document.body.getElementsByClassName('eq-idle');
   for (var i = 0, l = el.length; i < l; i++)
      el[i].classList.add('eq-fog');
};

/*------------------------------------------------------------------------
   static App.createIdlePane()
   Create idle pane.
------------------------------------------------------------------------*/

eq.c.App.createIdlePane = function(parent) {
   var el = document.createElement('DIV');
   el.className = 'eq-idle';
   el.style.zIndex = eq.Dom.zIndex(parent) + 2;
   parent.appendChild(el);
   return el;
};

/*------------------------------------------------------------------------
   App.stopIdleTimer()
   Stop application idle timer.
------------------------------------------------------------------------*/

eq.c.App.prototype.stopIdleTimer = function() {
   if (this.idleTimer !== undefined) {
      window.clearTimeout(this.idleTimer);
      this.idleTimer = undefined;
   }
};

/*------------------------------------------------------------------------
   App.makeOverlay() .closeOverlay()
   Overlay management.
------------------------------------------------------------------------*/

eq.c.App.prototype.makeOverlay = function(el, overlay, onClose) {
   if (this.overlay !== undefined)
      throw new Error("makeOverlay BUG: pending overlay id:" +
         this.overlay.el.id);
   if (document.body.querySelector('body>.eq-overlay'))
      throw new Error("makeOverlay BUG: orphaned overlay");
   this.overlay = {
      el : el,
      overlay : overlay,
      onClose : onClose
   };
   var b = document.body;
   overlay.classList.add('eq-overlay');
   overlay.style.zIndex = eq.Dom.zIndex(b) + 1;
   b.appendChild(overlay);
};

eq.c.App.prototype.closeOverlay = function(el, arg) {
   var od = this.overlay;
   if (od === undefined) {
      if (document.body.querySelector('body>.eq-overlay'))
         throw new Error("closeOverlay BUG: orphaned overlay");
      return;
   }
   var ov = od.overlay;
   if (el !== undefined && el !== od.el)
      throw new Error("closeOverlay BUG: wrong id:" + el.id +
         ", expected:" + od.el.id);
   if (!ov.classList.contains('eq-overlay'))
      throw new Error("closeOverlay BUG: no overlay id:" + od.el.id);
   this.overlay = undefined;
   od.onClose(od.el, ov, arg);
   ov.style.zIndex = '';
   ov.classList.remove('eq-overlay');
};

/*------------------------------------------------------------------------
   App.newDialog()
   Create new dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.newDialog = function(id) {
   var lm = this.layoutMode, dlg;
   if (lm !== undefined && (lm & eq.layoutMode.inline)
                        && !(lm & eq.layoutMode.dialog)) {
      // Single inline Dialog, forced by layout mode,
      // invalidate and close existing dialogs.
      this.dlg.forEach(eq.c.App.cbInvalidateCloseDlg, this);
      this.dlg.clear();
   }

   dlg = new eq.c.Dlg(id);
   this.dlg.set(id, dlg);
   return dlg;
};

eq.c.App.cbInvalidateCloseDlg = function(d) {
   this.api.postMessage({
      o : eq.ApiOp.invalidate,
      d : d.id
   });
   d.close();
};

/*------------------------------------------------------------------------
   App.closeDialog()
   Close dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.closeDialog = function(dlg) {
   this.dlg.delete(dlg.id);
   dlg.close();
};

/*------------------------------------------------------------------------
   App.topDialog()
   Get topmost dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.topDialog = function() {
   var wl, i, d;
   if ((wl = this.wl) !== undefined)
      for (i = wl.length; --i >= 0;)
         if ((d = wl[i].dlg) !== undefined)
            return d;

   return null;
};

/*------------------------------------------------------------------------
   App.setCurrDialog()
   Set current dialog, make it top and visible.
------------------------------------------------------------------------*/

eq.c.App.prototype.setCurrDialog = function(dlg) {
   if (dlg.w !== null)
      dlg.w.activate();

   dlg.visible();
   this.currDlg = dlg;

   // Set Dialog title.
   var t = dlg.title || eq.c.App.title;
   if (t !== document.title) {
      document.title = t;
      eq.app.updateSessionList();
   }
};

/*------------------------------------------------------------------------
   App.dialogsVisible()
   Make dialogs visible.
------------------------------------------------------------------------*/

eq.c.App.prototype.dialogsVisible = function() {
   this.dlg.forEach(eq.c.App.cbDialogsVisible);
};

eq.c.App.cbDialogsVisible = function(d) {
   d.visible();
};

/*========================================================================
   class Dlg extends Obj
   DLG dialog instance.
========================================================================*/

eq.c.Obj.subClass('Dlg', function(id) {
   eq.c.Obj.call(this);

   // Dialog id.
   this.id = id;

   // Window.
   this.w = null;

   // Root elements.
   this.root = [];

   // Element descriptors.
   this.eds = new Map();

   // Tab order.
   this.tabOrder = null;
});

/*------------------------------------------------------------------------
   Dlg.close()
   Close dialog instance.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.close = function() {
   var app = eq.c.App;

   // Reset current Dialog if necessary.
   this.resetCurrDlg();

   // Restore original page title if necessary.
   if (document.title === this.title) {
      var t = app.title;
      if (t !== document.title) {
         document.title = t;
         eq.app.updateSessionList();
      }
   }

   // Remove root elements.
   for (var i = this.root.length; --i >= 0;) {
      var el = this.root[i];
      app.removeRootListeners(el);
      if (!el.classList.contains('eq-dialog'))
         el.parentNode.removeChild(el);
      else {
         // Close Window.
         if (this.w === null)
            throw new Error("Dlg.close(" + this.id + ") BUG: no Window");
         this.w.close();
         this.w = null;
      }
   }
   this.root.length = 0;
   if (this.w !== null)
      throw new Error("Dlg.close(" + this.id + ") BUG: Window not closed");

   // Clear element descriptors.
   this.eds.clear();

   // Clear tab order.
   this.tabOrder = null;

   // Clear pending close request.
   this.pendingClose = undefined;

   // Delete properties.
   this.layout = undefined;
   this.title = undefined;
   this.fkey = undefined;
   this.ddo = undefined;
};

eq.c.Dlg.prototype.resetCurrDlg = function() {
   var app = eq.app;
   if (this === app.currDlg) {
      app.currDlg = null;
      app.focus = null;
      app.changed.clear();
      app.setIdle(this);
      this.ddo = undefined;
   }
};

/*------------------------------------------------------------------------
   Dlg.addRoot()
   Add dialog root element.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.addRoot = function(el) {

   // TODO: dlgAnchor

   var cl = el.classList, wc, lm, l, r;
   el.style.display = 'none';
   cl.add('eq-root');
   if (cl.contains('eq-dialog')) {
      if (this.w !== null)
         throw new Error("Dlg.addRoot(" + this.id + ") BUG: Window exists");
      wc = eq.c.Window;
      if (   (lm = eq.app.layoutMode) !== undefined
          && (lm & eq.layoutMode.inline) && !(lm & eq.layoutMode.dialog)) {
         // Inline Dialog, forced by layout mode.
         r = el;
      }
      else {
         lm = undefined;
         if (  (l = this.layout) === undefined
             || l.mode === eq.DialogMode.inline) {
            // Inline Dialog.
            r = el;
         }
         else {
            // Dialog in frame.
            r = wc.setupFrame(el);
         }
      }
      this.w = new wc(r, document.getElementById('eq-dialog'), this);
      if (lm !== undefined) {
         // Inline Dialog, forced by layout mode.
         cl.add('eq-inline');
         if (lm & eq.layoutMode.maximized)
            cl.add('eq-max');
         else
            cl.remove('eq-max');
      }
   }
   else
      throw new Error("Dlg.addRoot(" + this.id + ") BUG: not a Dialog");

   this.root.push(el);
   eq.c.App.addRootListeners(el);
};

/*------------------------------------------------------------------------
   Dlg.removeElement()
   Remove element from dialog.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.removeElement = function(el) {
   var r = this.root, eds = this.eds, i, empty, n;
   if ((i = r.indexOf(el)) !== -1) {
      // Root element.
      r.splice(i, 1);
      empty = r.length === 0;
   }

   if (!empty) {
      // Remove from element descriptors.
      n = el.querySelectorAll('[id]');
      for (i = n.length; --i >= 0;)
         eds.delete(n[i].id);
      eds.delete(el.id);
   }

   if (empty === undefined)
      el.parentNode.removeChild(el);
   else {
      // Root element.   
      eq.c.App.removeRootListeners(el);
      if (!el.classList.contains('eq-dialog'))
         el.parentNode.removeChild(el);
      else {
         // Close Window.
         if (this.w === null)
            throw new Error("Dlg.removeElement(" + this.id +
               ") BUG: no Window");
         this.w.close();
         this.w = null;
      }
   }

   return empty;
};

/*------------------------------------------------------------------------
   Dlg.visible()
   Make dialog visible.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.visible = function() {
   for (var ri = this.root.length; --ri >= 0;) {
      var rl = this.root[ri], st;
      if (rl.classList.contains('eq-dialog')) {
         if (this.w === null)
            throw new Error("Dlg.visible(" + this.id + ") BUG: no Window");

         // Validate existing panel elements,
         // relink hidden MenuBar and hidden/misplaced ToolBar(s).

         // Try to avoid flicker while relinking.
         if ((st = rl.style).display === 'none') {
            st.visibility = 'hidden';
            st.display = '';
         }
         if (   rl !== this.w.root
             && (st = this.w.root.style).display === 'none') {
            st.visibility = 'hidden';
            st.display = '';
         }

         var
            dc = rl.querySelector(
               '.eq-dialog>.eq-pane>.eq-container:not(.eq-panel)'),
            dp = dc.parentNode,
            el = rl.firstChild,
            tc = 'eq-top',
            cl, ch;
         while (el) {
            if ((cl = el.classList).contains('eq-panel')) {
               if (   cl.contains('eq-hidden')
                   || (cl.contains('eq-toolbar') && !cl.contains(tc))) {
                  var nx = el.nextSibling;
                  dc.appendChild(el);
                  cl.remove('eq-rt');
                  cl.remove('eq-panel');
                  cl.remove('eq-horz');
                  el = nx;
                  continue;
               }
            }
            else if (cl.contains('eq-pane')) {
               ch = el.firstChild;
               tc = 'eq-left';
               while (ch) {
                  if ((cl = ch.classList).contains('eq-panel')) {
                     if (   cl.contains('eq-hidden')
                         || (cl.contains('eq-toolbar') && !cl.contains(tc))) {
                        var nx = ch.nextSibling;
                        dc.appendChild(ch);
                        cl.remove('eq-rt');
                        cl.remove('eq-panel');
                        cl.remove('eq-vert');
                        ch = nx;
                        continue;
                     }
                  }
                  else if (cl.contains('eq-container'))
                     tc = 'eq-right';
                  ch = ch.nextSibling;
               }
               tc = 'eq-bottom';
            }
            el = el.nextSibling;
         }

         // MenuBar.
         if ((cl = (el = rl.firstChild).classList).contains('eq-idle'))
            cl = (el = el.nextSibling).classList;
         if (!cl.contains('eq-menubar')) {
            // Put first visible MenuBar into top-most panel.
            if (ch = dc.querySelector('.eq-menubar:not(.eq-hidden)')) {
               (cl = ch.classList).add('eq-rt');
               cl.add('eq-panel');
               rl.insertBefore(ch, el);
            }
         }

         // StatusBar.
         if ((cl = (el = rl.lastChild).classList).contains('eq-idle'))
            cl = (el = el.previousSibling).classList;
         if (!cl.contains('eq-statusbar')) {
            // Put first visible StatusBar into bottom-most panel.
            if (ch = dc.querySelector('.eq-statusbar:not(.eq-hidden)')) {
               (cl = ch.classList).add('eq-rt');
               cl.add('eq-panel');
               rl.appendChild(ch);
            }
         }

         // ToolBars.
         var n = dc.querySelectorAll('.eq-toolbar:not(.eq-hidden)');
         for (var i = 0, l = n.length; i < l; i++) {
            if ((cl = (el = n[i]).classList).contains('eq-top')) {
               cl.add('eq-rt');
               cl.add('eq-panel');
               cl.add('eq-horz');
               if ((ch = rl.firstChild).classList.contains('eq-idle'))
                  ch = ch.nextSibling;
               if (ch.classList.contains('eq-menubar'))
                  rl.insertBefore(el, ch.nextSibling);
               else
                  rl.insertBefore(el, ch);
            }
            else if (cl.contains('eq-bottom')) {
               cl.add('eq-rt');
               cl.add('eq-panel');
               cl.add('eq-horz');
               if ((ch = rl.lastChild).classList.contains('eq-idle'))
                  ch = ch.previousSibling;
               if (ch.classList.contains('eq-statusbar'))
                  rl.insertBefore(el, ch);
               else
                  rl.appendChild(el);
            }
            else if (cl.contains('eq-left')) {
               cl.add('eq-rt');
               cl.add('eq-panel');
               cl.add('eq-vert');
               dp.insertBefore(el, dp.firstChild);
            }
            else if (cl.contains('eq-right')) {
               cl.add('eq-rt');
               cl.add('eq-panel');
               cl.add('eq-vert');
               dp.appendChild(el);
            }
         }

         // Make Window visible on next frame,
         // set position and title if necessary.
         eq.Dom.onNextFrame(this, { cb : eq.c.Dlg.commitVisible, rl : rl });
      }
      else {
         // Make root element visible.
         rl.style.display = '';
      }
   }
};

eq.c.Dlg.commitVisible = function(d, qi) {
   var w = d.w, rl = qi.rl, r, l, cl, whenVisible;
   if (w === null)
      throw new Error("Dlg.commitVisible(" + d.id + ") BUG: no Window");
   if ((r = w.root) === undefined)
      throw new Error("Dlg.commitVisible(" + d.id +
         ") BUG: Window root undefined");
   whenVisible = { cb : eq.c.Dlg.whenVisible };
   if ((l = d.layout) === undefined) {
      rl.style.visibility = '';
      r.style.visibility = '';
   }
   else if (   (lm = eq.app.layoutMode) !== undefined
            && (lm & eq.layoutMode.inline) && !(lm & eq.layoutMode.dialog)) {
      // Inline Dialog, forced by layout mode.
      d.layout = undefined;
      rl.style.visibility = '';
   }
   else {
      d.layout = undefined;
      if (l.mode !== undefined) {
         if (l.mode === eq.DialogMode.inline) {
            // Inline Dialog.
            if (w.makeInline())
               r = w.root;
         }
         else {
            // Dialog in frame.
            if (w.makeFramed())
               r = w.root;
         }
      }
      if ((cl = r.classList).contains('eq-dialog'))
         rl.style.visibility = '';
      else {
         // Dialog in frame.
         w.setTitle(d.title ? d.title : "Eloquence");
         if (l.mode !== undefined)
            if (l.mode === eq.DialogMode.layout) {
               cl.add('eq-resize');
               if (l.state !== undefined)
                  if (l.state & eq.DialogState.maximized)
                     cl.add('eq-max');
                  else
                     cl.remove('eq-max');
            }
            else {
               cl.remove('eq-resize');
               cl.remove('eq-max');
            }
         if (rl.classList.contains('eq-rule'))
            cl.add('eq-close');
         else {
            cl.remove('eq-close');
            d.pendingClose = undefined;
         }
         rl.style.visibility = '';
         if (l.x !== undefined && l.y !== undefined) {
            w.whenVisible = whenVisible;
            whenVisible = undefined;
            w.setPos(l.x, l.y);
         }
         else
            r.style.visibility = '';
      }
   }
   if (whenVisible !== undefined)
      eq.Dom.onNextFrame(w, whenVisible);
};

eq.c.Dlg.whenVisible = function() {
   // Adjust Window list stack order if necessary.
   eq.c.Window.zOrderModified();
   // Restore focus.
   var app = eq.app, d, f, el, ed;
   if (   (d = app.currDlg)
       && (f = app.focus) !== null
       && (el = document.getElementById(f))
       && el.classList.contains('eq-focused')
       && (ed = d.eds.get(f)) !== undefined) {
      if ((cb = eq.FocusElement.get(ed.cb)) !== undefined)
         el = cb(el);
      eq.Dom.onNextFrame(el, { cb : eq.c.App.setElementFocus});
   }
};

/*------------------------------------------------------------------------
   Dlg.addControl()
   Add DLG control to document from DOM element template(s).
   See: dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.addControl = function(et, elParent, rootId) {
   var root;
   if (rootId === undefined && !(rootId = et.id))
      throw new Error("addControl: root id not set");

   while (true) {
      // Create element.
      var
         tg = et.tg,
         el = document.createElement(tg !== undefined ? tg : 'DIV');
      this.updateControl(rootId, el, et);
      if (root === undefined)
         root = el;

      // Add to DOM.
      if (elParent !== undefined)
         elParent.appendChild(el);
      elParent = el;

      // Nested template.
      if (et.et === undefined)
         break;
      et = et.et;
   }

   // Children.
   var ch = et.ch;
   if (ch !== undefined) {
      for (var i = 0, l = ch.length; i < l; i++) {
         var ct = ch[i], id = ct.id;
         if (id === '*')
            ct.id = (id = rootId) + '-' + (i+1);
         this.addControl(ct, elParent, id || rootId);
      }
   }

   return root;
};

/*------------------------------------------------------------------------
   Dlg.modifyControl()
   Modify DLG control from DOM element template(s).
   See: dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.modifyControl = function(rootId, el, et) {
   while (true) {
      var tg = et.tg;
      if (tg !== undefined && tg !== el.tagName) {
         // Tag modified, recreate element, retain 'eq-event' if set.
         var en = document.createElement(tg), cl = el.classList;
         if (cl && cl.contains('eq-event'))
            en.classList.add('eq-event');
         el.parentNode.replaceChild(en, el);
         el = en;
      }

      // Update element.
      this.updateControl(rootId, el, et);

      // Nested template?
      if (et.et === undefined)
         break;
      // Yes, use first child which is not 'runtime-created'.
      var cn = el.childNodes;
      el = undefined;
      et = et.et;
      for (var i = 0, l = cn.length; i < l; i++) {
         var ci = cn[i];
         if (!ci.classList.contains('eq-rt')) {
            el = ci;
            break;
         }
      }
      if (el === undefined)
         throw new Error("modifyControl: id '" + rootId +
            "' nested template, no child");
   }

   // Children.
   var ch = et.ch;
   if (ch !== undefined) {
      var cn = el.childNodes;
      for (var i = 0, l = ch.length; i < l; i++) {
         var ci = ch[i];
         if (i >= cn.length) {
            var tg = ci.tg;
            el.appendChild(document.createElement(
               tg !== undefined ? tg : 'DIV'));
         }
         this.modifyControl(rootId, cn[i], ci);
      }
      while (cn.length > l)
         el.removeChild(cn[l]);
   }
};

/*------------------------------------------------------------------------
   Dlg.updateControl()
   Update DLG control from DOM element template.
   See: dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.updateControl = function(rootId, el, et) {
   for (var a in et)
      switch (a) {
         case 'id': {
            // id
            var
               id = el.id,
               nid = et.id,
               eds, ed;
            if (id !== nid) {
               if (id) {
                  if ((ed = (eds = this.eds).get(id)) === undefined)
                     throw new Error("BUG: updateControl id:" + id +
                        " not registered");
                  eds.delete(id);
                  eds.set(nid, ed);
               }
               el.id = nid;
            }
            break;
         }

         case 'cb': {
            // Callback.
            var
               id = el.id,
               eds = this.eds,
               ed;
            if (!id)
               throw new Error("BUG: updateControl undefined callback id");
            if ((ed = eds.get(id)) !== undefined)
               ed.cb = et.cb;
            else
               eds.set(id, { cb : et.cb });
            break;
         }

         case 'hc':
            // Text or HTML content.
            eq.Dom.setContent(el, et.hc);
            break;

         case 'hca':
            // Text or HTML content w/ optional accelerator.
            eq.Dom.setContent(el, et.hca, true);
            break;

         case 'ec':
            // EditText content.
            el.value = et.ec;
            if (el.tagName === 'TEXTAREA')
               el.textContent = et.ec;
            break;

         case 'at': {
            // Attributes.
            for (var at in et.at) {
               var a = et.at[at];
               if (a !== null)
                  el.setAttribute(at, a);
               else
                  el.removeAttribute(at);
            }
            break;
         }

         case 'clr': {
            // Remove class names.
            for (var i = 0, l = et.clr.length; i < l; i++)
               el.classList.remove(et.clr[i]);
            break;
         }

         case 'cla': {
            // Add class names.
            for (var i = 0, l = et.cla.length; i < l; i++)
               el.classList.add(et.cla[i]);
            break;
         }

         case 'st': {
            // Styles.
            for (var st in et.st) {
               var s = et.st[st];
               el.style.setProperty(st, s !== null ? s : '');
            }
            break;
         }

         case 'rt': {
            // Value by request tag, invoke setAttr callback.
            var
               ed = this.eds.get(rootId),
               cb;
            if (ed === undefined)
               throw new Error("BUG: updateControl id:" + rootId +
                  " not registered");
            if ((cb = eq.setAttr.get(ed.cb)) === undefined)
               throw new Error("BUG: updateControl id:" + rootId +
                  " setAttr callback undefined");
            cb(this, rootId, el, et.rt);
            break;
         }

         case 'ev': {
            // Events.
            var
               ed = this.eds.get(rootId),
               ev = et.ev;
            if (ed === undefined)
               throw new Error("BUG: updateControl id:" + rootId +
                  " not registered");
            if (ev.a !== undefined)
               ed.a = ev.a !== null ? ev.a : undefined;
            if (ev.c !== undefined)
               ed.c = ev.c !== null ? ev.c : undefined;
            if (ev.m !== undefined)
               ed.m = ev.m !== null ? ev.m : undefined;
            if (ev.f !== undefined)
               ed.f = ev.f !== null ? ev.f : undefined;
            if (ed.a !== undefined || ed.c !== undefined)
               el.classList.add('eq-event');
            else
               el.classList.remove('eq-event');
            break;
         }

         case 'dlg': {
            var d = et.dlg;
            if (d.l !== undefined)
               this.layout = d.l ? d.l : undefined;
            if (d.t !== undefined)
               this.title = d.t ? d.t : undefined;
            if (d.f !== undefined)
               this.fkey = d.f ? d.f : undefined;
            if (d.d !== undefined) {
               if (d.d)
                  this.ddo = d.d;
               else if (this.ddo !== undefined) {
                  this.ddo = undefined;
                  this.resetCurrDlg();
               }
            }
            break;
         }
      }
};

/*========================================================================
   Start main thread
========================================================================*/

eq.app = new eq.c.App();
if (!eq.app.redirected) {
   // Install global event listeners.
   window.addEventListener('keydown', eq.c.App.onGlobalKeyDown, true);
   window.addEventListener('keyup', eq.c.App.onGlobalKeyUp, true);
   window.addEventListener('click', eq.c.App.onGlobalMouseDownOrClick, true);
   window.addEventListener('mousedown', eq.c.App.onGlobalMouseDownOrClick, true);
   window.addEventListener('mouseup', eq.c.App.onGlobalMouseUp, true);
   window.addEventListener('resize', eq.c.App.onGlobalResize, true);
   window.addEventListener('blur', eq.c.App.onGlobalBlur, true);
   window.addEventListener('pagehide', eq.c.App.onGlobalPageHide, true);
   window.addEventListener('storage', eq.c.App.onStorageModified);

   // Start API thread.
   eq.c.App.apiStart();

   // Continue initialization when DOM is ready.
   if (document.readyState === 'loading')
      window.addEventListener('DOMContentLoaded', eq.c.App.domReady);
   else
      eq.c.App.domReady();
}

